//
//  AABBtree.cpp
//  ParchmentFlattener
//
//  Copyright (c) 2014 Kazim Pal.
//  Released under the MIT License.
//
#include "AABBtree.h"
#include "IGLMesh.h"
#include <QSet>

#define EPSILON 0.000001
#define NOT_INTERSECT 10000000

AABBNode::AABBNode(void)
{
    left = 0;
    right = 0;
}

AABBNode::AABBNode(QVector<int> faceList, AABBTreeInfo &treeInfo, int depth, IGLMesh *mesh)
{
    left = 0;
    right = 0;
    BuildTree(faceList, treeInfo, depth, mesh);
}

AABBNode::~AABBNode(void)
{
    if(left)
        delete left;
    if(right)
        delete right;
}

bool AABBNode::BoxIntersectsRay(Vector3d O, Vector3d D, double t0, double t1)
{
    double tmin, tmax, tymin, tymax, tzmin, tzmax;

    if (D(0) >= 0) {
        tmin = (p_min(0) - O(0)) / D(0);
        tmax = (p_max(0) - O(0)) / D(0);
    }
    else {
        tmin = (p_max(0) - O(0)) / D(0);
        tmax = (p_min(0) - O(0)) / D(0);
    }

    if (D(1) >= 0) {
        tymin = (p_min(1) - O(1)) / D(1);
        tymax = (p_max(1) - O(1)) / D(1);
    }
    else {
        tymin = (p_max(1) - O(1)) / D(1);
        tymax = (p_min(1) - O(1)) / D(1);
    }

    if ( (tmin > tymax) || (tymin > tmax) )
        return false;

    if (tymin > tmin)
        tmin = tymin;

    if (tymax < tmax)
        tmax = tymax;

    if (D(2) >= 0) {
        tzmin = (p_min(2) - O(2)) / D(2);
        tzmax = (p_max(2) - O(2)) / D(2);
    }
    else {
        tzmin = (p_max(2) - O(2)) / D(2);
        tzmax = (p_min(2) - O(2)) / D(2);
    }

    if ( (tmin > tzmax) || (tzmin > tmax) )
        return false;

    if (tzmin > tmin)
        tmin = tzmin;

    if (tzmax < tmax)
        tmax = tzmax;

    return ( (tmin < t1) && (tmax > t0) );
}

bool AABBNode::IntersectTriangleWithRay(Vector3d O, Vector3d D, int face_index, IGLMesh *mesh, double *t, double *u, double *v)
{
    Vector3d edge1, edge2, tvec, pvec, qvec;
    double det, inv_det;

    int v0 = mesh->F(0,node_faces[face_index]);
    int v1 = mesh->F(1,node_faces[face_index]);
    int v2 = mesh->F(2,node_faces[face_index]);

    Vector3d vert0 = mesh->flatV.col(v0);
    Vector3d vert1 = mesh->flatV.col(v1);
    Vector3d vert2 = mesh->flatV.col(v2);

    edge1 = vert1 - vert0;
    edge2 = vert2 - vert0;

    pvec = D.cross(edge2);

    det = edge1.dot(pvec);

    if(det > -EPSILON && det < EPSILON)
        return false;
    inv_det = 1.0 / det;

    tvec = O - vert0;

    *u = tvec.dot(pvec) * inv_det;
    if(*u < 0.0 || *u > 1.0)
        return false;

    qvec = tvec.cross(edge1);

    *v = D.dot(qvec) * inv_det;
    if(*v < 0.0 || *v > 1.0)
        return false;

    double _w = 1 - *u - *v;
    if(_w < 0.0 || _w > 1.0)
        return false;

    *t = edge2.dot(qvec) * inv_det;
    return true;
}

void AABBNode::IntersectRay(Vector3d O, Vector3d D, IGLMesh *mesh, double *t, int *k, double *u, double *v)
{
    if(BoxIntersectsRay(O,D,0,100))
    {
        if(isLeaf)
        {
            double tBest = NOT_INTERSECT;
            int triBest = -1;
            double uBest;
            double vBest;
            for(int i = 0; i < numFaces; ++i)
            {
                double t = 0.0;
                double uf = 0;
                double vf = 0;
                if(IntersectTriangleWithRay(O,D,i,mesh,&t,&uf,&vf))
                {
                    if(t < tBest)
                    {
                        tBest = t;
                        uBest = uf;
                        vBest = vf;
                        triBest = node_faces[i];
                    }
                }
            }
            if(triBest > -1)
            {
                *t = tBest;
                *k = triBest;
                *u = uBest;
                *v = vBest;
            }
            else
            {
                *t = NOT_INTERSECT;
                *k = -1;
                *u = -1;
                *v = -1;
            }
        }
        else
        {
            double tleft,uleft,vleft,tright,uright,vright;
            int kleft,kright;

            left->IntersectRay(O,D,mesh,&tleft,&kleft,&uleft,&vleft);
            right->IntersectRay(O,D,mesh,&tright,&kright,&uright,&vright);
            if(tleft < tright)
            {
                *t = tleft;
                *k = kleft;
                *u = uleft;
                *v = vleft;
            }
            else
            {
                *t = tright;
                *k = kright;
                *u = uright;
                *v = vright;
            }
        }
    }
    else
    {
        *t = NOT_INTERSECT;
        *k = -1;
        *u = -1;
        *v = -1;
    }
}

int AABBNode::FindBestAxis(QVector<int> faceList, IGLMesh *mesh)
{
    numFaces = faceList.size();

    int iAxis=0;
    int iAxisResult[3];

    Vector3d center = (p_min + p_max)/2;

    for (iAxis=0;iAxis<3;iAxis++)
    {
        int left=0,right=0;

        double axisCenter = center(0);
        if(iAxis == 1)
            axisCenter = center(1);
        if(iAxis == 2)
            axisCenter = center(2);

        for(int i=0; i<numFaces; ++i)
        {
            Vector3d p0 = mesh->flatV.col(mesh->F(0,faceList[i]));
            Vector3d p1 = mesh->flatV.col(mesh->F(1,faceList[i]));
            Vector3d p2 = mesh->flatV.col(mesh->F(2,faceList[i]));
            Vector3d faceCenter = (p0 + p1 + p2) / 3.0f;

            double faceCenterD = faceCenter(0);
            if(iAxis == 1)
                faceCenterD = faceCenter(1);
            if(iAxis == 2)
                faceCenterD = faceCenter(2);

            if (faceCenterD <= axisCenter)
            {
                left++;
            }
            else
            {
                right++;
            }
        }

        iAxisResult[iAxis] = abs(left-right);
    }

    int index=0;
    int result=iAxisResult[0];
    for (int i=1;i<3;i++)
    {
        if (iAxisResult[i] < result)
        {
            result = iAxisResult[i];
            index=i;
        }
    }

    return index;
}

void AABBNode::GetFacesWithinRadius(Vector3d p, double r2, QSet<int> *fs, IGLMesh *mesh)
{
    if(isLeaf)
    {
        for(int k = 0; k < node_faces.size(); ++k)
        {
            int findex = node_faces[k];

            Vector3d v0 = mesh->flatV.col(mesh->F(0,findex));
            Vector3d v1 = mesh->flatV.col(mesh->F(1,findex));
            Vector3d v2 = mesh->flatV.col(mesh->F(2,findex));

            double d2_0 = (v0-p).norm(); d2_0 = d2_0*d2_0;
            double d2_1 = (v1-p).norm(); d2_1 = d2_1*d2_1;
            double d2_2 = (v2-p).norm(); d2_2 = d2_2*d2_2;

            if(d2_0 < r2 && d2_1 < r2 && d2_2 < r2)
                fs->insert(findex);
        }
    }
    else
    {
        left->GetFacesWithinRadius(p,r2,fs,mesh);
        right->GetFacesWithinRadius(p,r2,fs,mesh);
    }
}

double AABBNode::SqDistToPoint(Vector3d p)
{
    double sqDist = 0;

    // x
    double px = p(0);
    if(px < p_min(0))
        sqDist += (p_min(0) - px)*(p_min(0) - px);
    if(px > p_max(0))
        sqDist += (px - p_max(0))*(px - p_max(0));

    // y
    double py = p(1);
    if(py < p_min(1))
        sqDist += (p_min(1) - py)*(p_min(1) - py);
    if(py > p_max(1))
        sqDist += (py - p_max(1))*(py - p_max(1));

    // z
    double pz = p(2);
    if(pz < p_min(2))
        sqDist += (p_min(2) - pz)*(p_min(2) - pz);
    if(pz > p_max(2))
        sqDist += (pz - p_max(2))*(pz - p_max(2));

    return sqDist;
}

void AABBNode::BuildTree(QVector<int> faceList, AABBTreeInfo &treeInfo, int depth, IGLMesh *mesh)
{
    numFaces = faceList.size();

    // Construct bounding box
    p_min = Vector3d(9999,9999,9999);
    p_max = -p_min;
    for(int f = 0; f < faceList.size(); ++f)
    {
        Vector3d p0 = mesh->flatV.col(mesh->F(0,faceList[f]));
        Vector3d p1 = mesh->flatV.col(mesh->F(1,faceList[f]));
        Vector3d p2 = mesh->flatV.col(mesh->F(2,faceList[f]));

        double fminx = qMin(qMin(p0(0),p1(0)),p2(0));
        double fminy = qMin(qMin(p0(1),p1(1)),p2(1));
        double fminz = qMin(qMin(p0(2),p1(2)),p2(2));

        double fmaxx = qMax(qMax(p0(0),p1(0)),p2(0));
        double fmaxy = qMax(qMax(p0(1),p1(1)),p2(1));
        double fmaxz = qMax(qMax(p0(2),p1(2)),p2(2));

        p_min(0) = (qMin((double)p_min(0),fminx));
        p_min(1) = (qMin((double)p_min(1),fminy));
        p_min(2) = (qMin((double)p_min(2),fminz));

        p_max(0) = (qMax((double)p_max(0),fmaxx));
        p_max(1) = (qMax((double)p_max(1),fmaxy));
        p_max(2) = (qMax((double)p_max(2),fmaxz));
    }

    p_min(2) = -0.1;
    p_max(2) = 0.1;

    if (depth+1 > treeInfo.curr_max_depth)
        treeInfo.curr_max_depth=depth+1;

    bool bMakeChildren=false;

    if (numFaces > treeInfo.min_faces && depth < treeInfo.max_tree_depth)
    {
        bMakeChildren=true;
    }

    if (bMakeChildren)
    {
        isLeaf = false;
        int iAxis = FindBestAxis(faceList, mesh);

        //get center of the box for longest axis
        Vector3d center = (p_min + p_max)/2.0;

        double axisCenter = center(0);
        if(iAxis == 1)
            axisCenter = center(1);
        if(iAxis == 2)
            axisCenter = center(2);

        QVector<int> leftSide;
        QVector<int> rightSide;

        int leftCount=0,rightCount=0;

        for(int i=0; i<numFaces; ++i)
        {
            Vector3d p0 = mesh->flatV.col(mesh->F(0,faceList[i]));
            Vector3d p1 = mesh->flatV.col(mesh->F(1,faceList[i]));
            Vector3d p2 = mesh->flatV.col(mesh->F(2,faceList[i]));
            Vector3d faceCenter = (p0 + p1 + p2) / 3.0f;

            double faceCenterD = faceCenter(0);
            if(iAxis == 1)
                faceCenterD = faceCenter(1);
            if(iAxis == 2)
                faceCenterD = faceCenter(2);

            if (faceCenterD <= axisCenter)
            {
                //Store the verts to the left.
                leftSide.push_back(faceList[i]);
                leftCount++;
            }
            else
            {
                //Store the verts to the right.
                rightSide.push_back(faceList[i]);
                rightCount++;
            }
        }

        if (leftCount == 0 || rightCount == 0)
        {
            leftSide.clear();
            rightSide.clear();

            int leftMaxIndex = numFaces / 2;

            for(int i=0;i < numFaces;i++)
            {
                if (i<leftMaxIndex)
                {
                    //left node
                    leftSide.push_back(faceList[i]);
                }
                else
                {
                    rightSide.push_back(faceList[i]);
                }
            }
        }

        if (leftSide.size() > 0 && rightSide.size() > 0)
        {
            left = new AABBNode(leftSide,treeInfo, depth+1,mesh);
            right = new AABBNode(rightSide,treeInfo, depth+1,mesh);
        }
        else
        {
            bMakeChildren=false;
        }
    }

    if (!bMakeChildren)
    {
        node_faces = faceList;
        isLeaf = true;
    }
}

AABBTree::AABBTree(void)
{
    theMesh = 0;
    root = 0;
}

AABBTree::AABBTree(IGLMesh *mesh)
{
    theMesh = mesh;
    mesh->aabbtree = this;
    root = 0;
}

AABBTree::~AABBTree(void)
{
    if(root)
        delete root;
}

void AABBTree::rebuildTree()
{
    qDebug() << "Rebuilding aabbtree";
    if(root)
    {
        delete root;
        root = 0;
    }
    AABBTreeInfo treeInfo;
    treeInfo.max_tree_depth = 1000;
    treeInfo.min_faces = 5;
    treeInfo.curr_max_depth = 0;
    QVector<int> fs = QVector<int>();
    for(int k = 0; k < theMesh->F.cols(); ++k)
    {
        fs.push_back(k);
    }
    root = new AABBNode(fs,treeInfo,0,theMesh);
    qDebug() << "Rebuilt aabbtree";
}

void AABBTree::IntersectRay(Vector3d O, Vector3d D, double *t, int *k, double *u, double *v)
{
    if(root == 0)
    {
        *k = -1;
        return;
    }
    else
    {
        root->IntersectRay(O,D,theMesh,t,k,u,v);
    }
}

void AABBTree::GetFacesWithinRadius(Vector3d p, double r2, QSet<int> *fs)
{
    root->GetFacesWithinRadius(p,r2,fs,theMesh);
}
