//
//  AutomationEngine.cpp
//  ParchmentFlattener
//
//  Copyright (c) 2014 Kazim Pal.
//  Released under the MIT License.
//
#include "AutomationEngine.h"
#include <opencv/highgui.h>
#include <ParchmentFlattener.h>

#define DEBUG_IMAGES false

AutomationEngine::AutomationEngine()
{
}

AutomationEngine::~AutomationEngine()
{
}

Mat derivativeOfGaussian(int size, double sigma, int width)
{
    Mat x = Mat(size,width,CV_64F);
    int rad = size/2;
    double sum = 0;
    for(int i = 0; i < size; ++i)
    {
        double xLin = i - rad;
        double xGauss = -xLin * exp( ((-xLin)*(-xLin)) / (2*sigma*sigma) );
        for(int j = 0; j < width; ++ j)
        {
            x.at<double>(i,j) = xGauss;
            sum += xGauss;
        }
    }
    for(int i = 0; i < size; ++i)
    {
        for(int j = 0; j < width; ++ j)
        {
            x.at<double>(i,j) = x.at<double>(i,j) / sum;
        }
    }
    return x;
}

Mat derivativeOfGaussianHoriz(int size, double sigma, int height)
{
    Mat x = Mat(height,size,CV_64F);
    int rad = size/2;
    double sum = 0;
    for(int i = 0; i < size; ++i)
    {
        double xLin = i - rad;
        double xGauss = -xLin * exp( ((-xLin)*(-xLin)) / (2*sigma*sigma) );
        for(int j = 0; j < height; ++ j)
        {
            x.at<double>(j,i) = xGauss;
            sum += xGauss;
        }
    }
    for(int i = 0; i < size; ++i)
    {
        for(int j = 0; j < height; ++ j)
        {
            x.at<double>(j,i) = x.at<double>(j,i) / sum;
        }
    }
    return x;
}

Vector2i AutomationEngine::localizeLineInImage(QImage im)
{
    //qDebug() << "localize line";
    cv::Mat mat(im.height(), im.width(), CV_8UC4, im.bits(), im.bytesPerLine());
    cv::Mat Temp;
    mat.convertTo(Temp,CV_8U);
    cv::Mat greyIm;
    cv::cvtColor(Temp, greyIm, CV_BGRA2GRAY);

    Mat dIm = greyIm.clone();

    //Mat sumVec = Mat::zeros(greyIm.rows,1,CV_32SC1);
    //cv::reduce(greyIm,sumVec,1,CV_REDUCE_SUM,CV_32SC1);
    //cv::normalize(sumVec, sumVec, 0, 255, NORM_MINMAX, CV_32SC1);
    //imwrite("sumim.jpg",sumVec);

    Mat wideKernel;
    if(ParchmentFlattener::verticalLines)
    {
        int kernelHeight = im.height();
        wideKernel = derivativeOfGaussianHoriz(50,15,kernelHeight);
    }
    else
    {
        int kernelWidth = im.width();
        wideKernel = derivativeOfGaussian(80,30,kernelWidth);
    }

    filter2D(greyIm,dIm,5,wideKernel);

    if(ParchmentFlattener::verticalLines)
    {
        dIm = cv::abs(dIm);
        if(DEBUG_IMAGES)
        {
            imwrite("dim.jpg",dIm);
        }
    }

    cv::normalize(dIm, dIm, 0, 255, NORM_MINMAX, CV_32SC1);

    int x, y;

    if(ParchmentFlattener::verticalLines)
    {
        x = im.width()/2;
        y = im.height()/2 + 1;

        Mat dImRow = dIm.row(y);

        bool finished = false;
        while(!finished)
        {
            int xleft = x-1;
            int xright = x+1;

            int valx = dImRow.at<int>(0,x);
            int valright = dImRow.at<int>(0,xright);
            int valleft = dImRow.at<int>(0,xleft);

            if(valx >= valright && valx >= valleft)
            {
                finished = true;
            }
            else
            {
                if(valright > valleft)
                {
                    x = xright;
                }
                else
                {
                    x = xleft;
                }
            }
        }
        if(DEBUG_IMAGES)
        {
            for(int i = 0; i < greyIm.rows; ++i)
            {
                greyIm.at<int>(i,x) = 0;
            }

            for(int i = 0; i < dIm.rows; ++i)
            {
                for(int j = 0; j < dIm.cols; ++j)
                {
                    dIm.at<int>(i,j) = dImRow.at<int>(0,j);
                }
            }
            imwrite("mainim.jpg",greyIm);
            imwrite("gradim.jpg",dIm);
        }
    }
    else
    {
        x = im.width()/2 + 1;
        y = im.height()/2;


        Mat dImCol = dIm.col(x);

        bool finished = false;
        while(!finished)
        {
            int yup = y-1;
            int ydown = y+1;

            int valy = dImCol.at<int>(y,0);
            int valup = dImCol.at<int>(yup,0);
            int valdown = dImCol.at<int>(ydown,0);

            if(valy <= valup && valy <= valdown)
            {
                finished = true;
            }
            else
            {
                if(valup < valdown)
                {
                    y = yup;
                }
                else
                {
                    y = ydown;
                }
            }
        }
        if(DEBUG_IMAGES)
        {
            for(int i = 0; i < greyIm.cols; ++i)
            {
                greyIm.at<int>(y,i) = 0;
            }

            for(int i = 0; i < dIm.rows; ++i)
            {
                for(int j = 0; j < dIm.cols; ++j)
                {
                    dIm.at<int>(i,j) = dImCol.at<int>(i,0);
                }
            }
            imwrite("mainim.jpg",greyIm);
            imwrite("gradim.jpg",dIm);
        }
    }

    mat.release();
    Temp.release();
    greyIm.release();
    dIm.release();
    wideKernel.release();

    return Vector2i(x,y);
}
