//
//  IGLMesh.cpp
//  ParchmentFlattener
//
//  Copyright (c) 2014 Kazim Pal.
//  Released under the MIT License.
//
#ifndef IGL_HEADER_ONLY
#  define IGL_HEADER_ONLY
#endif
#include "IGLMesh.h"
#include <igl/readOBJ.h>
#include <igl/per_vertex_normals.h>
#include <igl/adjacency_list.h>
#include <igl/gradMat.h>
#include <igl/axis_angle_to_quat.h>
#include <igl/quat_to_mat.h>
#ifdef TARGET_OS_MAC
#include <GL/glu.h>
#else
#include <GL/glu.h>
#endif
#include <qdebug.h>
#include <vector>
#include "qfile.h"
#include "AABBtree.h"
#include "ParchmentFlattener.h"
#include <qmath.h>
#include "FlatteningEngine.h"
#include "NRosy/NRosyField.h"
#include <Eigen/SparseQR>
#include <Eigen/SparseLU>
#include <igl/cotmatrix.h>
#include <igl/tt.h>
#include <QQueue>

IGLMesh::IGLMesh(void)
{
    selected_vertex = -1;
    C = Vector3d(0,0,0);
    flattened = false;
    selectedLineConstraint = -1;
    selectedLineConstraintPoint = -1;
    selectedScaleConstraint = -1;

    aLineConstraintWasMoved = false;
    hover_k = 0;
    hover_u = 0;
    hover_v = 0;
    renderCrossField = true;
    smoothIsPaused = true;
    flatteningEnabled = false;
    flatteningLocked = true;

    needsFlattening = false;
    needsSmoothing = true;

    isAligned = false;

    flatteningEngine = new FlatteningEngine();

    aabbtree = 0;
}


IGLMesh::~IGLMesh(void)
{
    delete flatteningEngine;
    delete aabbtree;
    V.resize(0,0);
    F.resize(0,0);
    VN.resize(0,0);
    FN.resize(0,0);
    TC.resize(0,0);
    FTC.resize(0,0);
}

void IGLMesh::unSelectVertex()
{
    selected_vertex = -1;
}

void IGLMesh::saveFlatVertices(QString filename)
{
    QFile file(filename);
    file.open(QIODevice::WriteOnly | QIODevice::Text);
    QTextStream out(&file);
    for(int i = 0; i < flatV.cols(); ++i)
    {
        out << "v " << QString::number(flatV(0,i)) << " " << QString::number(flatV(1,i)) << " " << QString::number(flatV(2,i)) << "\n";
    }
    file.close();
}

void IGLMesh::saveFlatMesh(QString filename)
{
    QFile file(filename);
    file.open(QIODevice::WriteOnly | QIODevice::Text);
    QTextStream out(&file);
    QStringList tokens = ParchmentFlattener::meshFilename.split("/",QString::SkipEmptyParts);
    QString mtllib = tokens.back();
    mtllib.append(".mtl");
    out << "mtllib " << mtllib << "\n";

    for(int i = 0; i < flatV.cols(); ++i)
    {
        out << "v " << QString::number(flatV(0,i)) << " " << QString::number(flatV(1,i)) << " " << QString::number(flatV(2,i)) << "\n";
    }
    for(int i = 0; i < TC.rows(); ++i)
    {
        out << "vt " << QString::number(TC(i,0)) << " " << QString::number(TC(i,1)) << "\n";
    }

    int currentMtl = -1;
    qDebug() << "Fcols " << F.cols();
    for(int i = 0; i < F.cols(); ++i)
    {
        int v0 = F(0,i) + 1;
        int v1 = F(1,i) + 1;
        int v2 = F(2,i) + 1;

        int uv0 = FTC(i,0) + 1;
        int uv1 = FTC(i,1) + 1;
        int uv2 = FTC(i,2) + 1;

        int mtl = Materials(i);
        if(mtl != currentMtl)
        {
            out << "g Texture_" << QString::number(mtl) << "\n";
            out << "usemtl Texture_" << QString::number(mtl) << "\n";
            currentMtl = mtl;
        }
        out << "f " << QString::number(v0) << "/" << QString::number(uv0) << " " << QString::number(v1) << "/" << QString::number(uv1) << " " << QString::number(v2) << "/" << QString::number(uv2) << "\n";
    }
    file.close();
}

void IGLMesh::loadFlatVertices(QString filename)
{
    QFile file(filename);
    file.open(QIODevice::ReadOnly);
    QTextStream in(&file);

    MatrixXd Vinitload;
    Vinitload.setZero(V.rows(),V.cols());
    int i = 0;
    while(!in.atEnd())
    {
        QString line = in.readLine();
        QStringList fields = line.split(" ");
        if(fields.size() >= 4 && fields[0] == "v")
        {
            double x = fields[1].toDouble();
            double y = fields[2].toDouble();
            double z = fields[3].toDouble();
            Vinitload(0,i) = x;
            Vinitload(1,i) = y;
            Vinitload(2,i) = z;
            ++i;
        }
    }
    setFlatVertexPositions(Vinitload);
    setInitialVertexPositions(Vinitload);
    flattened = true;
    reGatherCurrentConstraints();
    file.close();
}

void IGLMesh::loadConstraints(QString filename)
{
    QFile file(filename);
    file.open(QIODevice::ReadOnly);
    QTextStream in(&file);

    QString fileMeshName;
    fileMeshName = in.readLine();
    bool readingLine = false;

    while(!in.atEnd())
    {
        QString line = in.readLine();
        QStringList fields = line.split(" ");

        if(fields.size() > 0)
        {
            if(fields[0] == "lc" || fields[0] == "sc")
            {
                if(readingLine)
                {
                    lineConstraints.push_back(floatingLineConstraint);
                    int lc_id = lineConstraints.size() - 1;
                    computeSingleLineConstraintRightAndUpVectors(&lineConstraints[lc_id]);
                    for(int lcp = 0; lcp < lineConstraints[lc_id].size(); ++lcp)
                    {
                        int lcp_face = lineConstraints[lc_id].faces[lcp];
                        perFaceVertexConstraintList[lcp_face].push_back(QPair<int,int>(lc_id,lcp));
                    }
                }
            }

            if(fields[0] == "lc" )
            {
                readingLine = true;
                floatingLineConstraint.clear();
                if(fields[1] == "vert")
                    floatingLineConstraint.mIsVertical = true;
            }
            else if(fields[0] == "sc")
            {
                readingLine = false;
                floatingScaleConstraint.clear();

                line = in.readLine();
                fields = line.split(" ");
                int k = fields[0].toInt();
                double u = fields[1].toDouble();
                double v = fields[2].toDouble();
                double sx = fields[3].toDouble();
                double sy = fields[3].toDouble();
                if(fields.size() > 4)
                {
                    sy = fields[4].toDouble();
                }
                line = in.readLine();
                fields = line.split(" ");
                int k0 = fields[0].toInt();
                double u0 = fields[1].toDouble();
                double v0 = fields[2].toDouble();
                line = in.readLine();
                fields = line.split(" ");
                int k1 = fields[0].toInt();
                double u1 = fields[1].toDouble();
                double v1 = fields[2].toDouble();
                floatingScaleConstraint.setP0Fixed(k0,u0,v0);
                floatingScaleConstraint.setP1Fixed(k1,u1,v1);
                floatingScaleConstraint.setMidPoint(k,u,v);
                floatingScaleConstraint.scale = Vector2d(sx,sy);

                scaleConstraints2.push_back(floatingScaleConstraint);

                floatingScaleConstraint.clear();

                int sc_id = scaleConstraints2.size() - 1;
                perFaceScaleConstraintList[k].push_back(sc_id);
            }
            else if(fields[0] == "scp")
            {
                readingLine = false;
                floatingScaleConstraint.clear();

                line = in.readLine();
                fields = line.split(" ");
                double xmid = fields[0].toDouble();
                double ymid = fields[1].toDouble();
                double zmid = fields[2].toDouble();

                line = in.readLine();
                fields = line.split(" ");
                double xbot = fields[0].toDouble();
                double ybot = fields[1].toDouble();
                double zbot = fields[2].toDouble();

                line = in.readLine();
                fields = line.split(" ");
                double xtop = fields[0].toDouble();
                double ytop = fields[1].toDouble();
                double ztop = fields[2].toDouble();

                int k, k0, k1;
                double u, u0, u1;
                double v, v0, v1;
                double t;

                Vector3d D(0.5,0.5,0.5);
                D.normalize();

                Vector3d Omid(xmid,ymid,zmid);
                Vector3d Obot(xbot,ybot,zbot);
                Vector3d Otop(xtop,ytop,ztop);

                Vector3d vup = Otop - Omid;
                double sx = vup.norm() * 2;
                double sy = sx;

                Omid = Omid - 0.1 * D;
                Obot = Obot - 0.1 * D;
                Otop = Otop - 0.1 * D;

                intersectMeshWithRay(Omid,D,&t,&k,&u,&v);
                intersectMeshWithRay(Obot,D,&t,&k0,&u0,&v0);
                intersectMeshWithRay(Otop,D,&t,&k1,&u1,&v1);

                floatingScaleConstraint.setP0Fixed(k0,u0,v0);
                floatingScaleConstraint.setP1Fixed(k1,u1,v1);
                floatingScaleConstraint.setMidPoint(k,u,v);
                floatingScaleConstraint.scale = Vector2d(sx,sy);

                finishScaleConstraint(false);
            }
            else
            {
                if(readingLine)
                {
                    int k = fields[0].toInt();
                    double u = fields[1].toDouble();
                    double v = fields[2].toDouble();
                    floatingLineConstraint.addPoint(k,u,v);
                }
            }
        }
    }
    if(readingLine)
    {
        lineConstraints.push_back(floatingLineConstraint);
        int lc_id = lineConstraints.size() - 1;
        computeSingleLineConstraintRightAndUpVectors(&lineConstraints[lc_id]);
        for(int lcp = 0; lcp < lineConstraints[lc_id].size(); ++lcp)
        {
            int lcp_face = lineConstraints[lc_id].faces[lcp];
            perFaceVertexConstraintList[lcp_face].push_back(QPair<int,int>(lc_id,lcp));
        }
    }
    file.close();
    reGatherCurrentConstraints();
}

void IGLMesh::saveConstraints(QString filename)
{
    QFile file(filename);
    file.open(QIODevice::WriteOnly | QIODevice::Text);
    QTextStream out(&file);
    out << meshName << "\n";
    for(int i = 0; i < lineConstraints.size(); ++i)
    {
        if(!lineConstraints[i].isActive())
            continue;

        if(lineConstraints[i].mIsVertical)
            out << "lc vert \n";
        else
            out << "lc horiz \n";
        for(int j = 0; j < lineConstraints[i].size(); ++j)
        {
            int kj = lineConstraints[i].faces[j];
            double uj = lineConstraints[i].sts(0,j);
            double vj = lineConstraints[i].sts(1,j);
            out << QString::number(kj) << " " << QString::number(uj) << " " << QString::number(vj) << "\n";
        }
    }
    for(int i = 0; i < scaleConstraints2.size(); ++i)
    {
        if(!scaleConstraints2[i].isActive())
            continue;

        out << "sc \n";
        int ki = scaleConstraints2[i].face;
        double ui = scaleConstraints2[i].u;
        double vi = scaleConstraints2[i].v;
        double sxi = scaleConstraints2[i].scale(0);
        double syi = scaleConstraints2[i].scale(1);
        int ki0 = scaleConstraints2[i].face0;
        double ui0 = scaleConstraints2[i].u0;
        double vi0 = scaleConstraints2[i].v0;
        int ki1 = scaleConstraints2[i].face1;
        double ui1 = scaleConstraints2[i].u1;
        double vi1 = scaleConstraints2[i].v1;
        out << QString::number(ki) << " " << QString::number(ui) << " " << QString::number(vi) << " " << QString::number(sxi) << " " << QString::number(syi) << "\n";
        out << QString::number(ki0) << " " << QString::number(ui0) << " " << QString::number(vi0) << "\n";
        out << QString::number(ki1) << " " << QString::number(ui1) << " " << QString::number(vi1) << "\n";
    }
    file.close();
}

void IGLMesh::saveConstraintsAbs(QString filename)
{
    QFile file(filename);
    file.open(QIODevice::WriteOnly | QIODevice::Text);
    QTextStream out(&file);
    out << meshName << "\n";
    for(int i = 0; i < lineConstraints.size(); ++i)
    {
        if(!lineConstraints[i].isActive())
            continue;

        if(lineConstraints[i].mIsVertical)
            out << "lcp vert \n";
        else
            out << "lcp horiz \n";
        for(int j = 0; j < lineConstraints[i].size(); ++j)
        {
            int kj = lineConstraints[i].faces[j];
            double uj = lineConstraints[i].sts(0,j);
            double vj = lineConstraints[i].sts(1,j);
            Vector3d pj = getWrinkledSurfacePoint(kj,uj,vj);
            out << QString::number(pj(0)) << " " << QString::number(pj(1)) << " " << QString::number(pj(2)) << "\n";
        }
    }
    for(int i = 0; i < scaleConstraints2.size(); ++i)
    {
        if(!scaleConstraints2[i].isActive())
            continue;

        if(scaleConstraints2[i].face2 >= 0)
            out << "scp \n";
        else
            out << "scpa \n";

        int ki = scaleConstraints2[i].face;
        double ui = scaleConstraints2[i].u;
        double vi = scaleConstraints2[i].v;

        int ki0 = scaleConstraints2[i].face0;
        double ui0 = scaleConstraints2[i].u0;
        double vi0 = scaleConstraints2[i].v0;

        int ki1 = scaleConstraints2[i].face1;
        double ui1 = scaleConstraints2[i].u1;
        double vi1 = scaleConstraints2[i].v1;

        Vector3d pi = getWrinkledSurfacePoint(ki,ui,vi);
        Vector3d pi0 = getWrinkledSurfacePoint(ki0,ui0,vi0);
        Vector3d pi1 = getWrinkledSurfacePoint(ki1,ui1,vi1);

        out << QString::number(pi(0)) << " " << QString::number(pi(1)) << " " << QString::number(pi(2)) << "\n";
        out << QString::number(pi0(0)) << " " << QString::number(pi0(1)) << " " << QString::number(pi0(2)) << "\n";
        out << QString::number(pi1(0)) << " " << QString::number(pi1(1)) << " " << QString::number(pi1(2)) << "\n";

        if(scaleConstraints2[i].face2 >= 0)
        {
            int ki2 = scaleConstraints2[i].face2;
            double ui2 = scaleConstraints2[i].u2;
            double vi2 = scaleConstraints2[i].v2;

            Vector3d pi2 = getWrinkledSurfacePoint(ki2,ui2,vi2);
            out << QString::number(pi2(0)) << " " << QString::number(pi2(1)) << " " << QString::number(pi2(2)) << "\n";
        }
    }
    file.close();
}

void IGLMesh::loadTextureCoords(QString filename)
{
    MatrixXd Vfoo;
    MatrixXi Ffoo;
    MatrixXd VNfoo;
    MatrixXi FNfoo;
    igl::readOBJ(filename.toStdString(),Vfoo,Ffoo,VNfoo,FNfoo,TC,FTC);

    QString mtlfname = filename;
    mtlfname.append(".mtl");
    loadMaterials(filename,mtlfname);
}

void IGLMesh::loadMesh(QString filename)
{
    V.resize(0,0);
    F.resize(0,0);
    VN.resize(0,0);
    FN.resize(0,0);
    TC.resize(0,0);
    FTC.resize(0,0);

    igl::readOBJ(filename.toStdString(),V,F,VN,FN,TC,FTC);
    V.transposeInPlace();
    Vuntouched = V;
    F.transposeInPlace();
    computeNormals();
    //igl::adjacency_list(F,adj);
    getIncidentFaces();
    C = Vector3d(0,0,0);
    for(int k = 0; k < V.cols(); ++k) {
        Vector3d R = V.col(k);
		C = C + R;
	}
    C = C / V.cols();
    centerMesh();
    orientMesh();

    vectorConstraints = MatrixXd::Zero(V.cols(),V.cols());
    isVectorConstrained = VectorXi::Zero(V.cols());
    scaleConstraints = VectorXd::Zero(V.cols());
    isScaleConstrained = VectorXi::Zero(V.cols());
    isFaceVectorConstrained = VectorXi::Zero(F.cols());
    //contributesToSmoothing = VectorXi::Zero(V.cols());

    QString mtlfname(filename);
    mtlfname.append(".mtl");
    loadMaterials(filename,mtlfname);

    perVertexRightVectorsW = MatrixXd::Zero(3,V.cols());
    perVertexUpVectorsW = MatrixXd::Zero(3,V.cols());    
    //perVertexScaleW.setOnes(2,V.cols());
    perVertexScaleW.setConstant(2,V.cols(),0.1);

    perVertexRightVectorsF = MatrixXd::Zero(3,V.cols());
    perVertexUpVectorsF = MatrixXd::Zero(3,V.cols());
    //perVertexScaleF.setOnes(2,V.cols());
    perVertexScaleF.setConstant(2,V.cols(),0.1);

    perFaceRightVectorsW.setZero(3,F.cols());
    perFaceUpVectorsW.setZero(3,F.cols());
    //perFaceScaleW.setOnes(2,F.cols());
    perFaceScaleW.setConstant(2,F.cols(),0.1);

    perFaceRightVectorsF.setZero(3,F.cols());
    perFaceUpVectorsF.setZero(3,F.cols());
    //perFaceScaleF.setOnes(2,F.cols());
    perFaceScaleF.setConstant(2,F.cols(),0.1);

    if(aabbtree)
        delete aabbtree;
    aabbtree = new AABBTree(this);

}

void IGLMesh::startSmoothing()
{
    smoothingAllowed = false;
    smoothThread = new QThread;
    SmoothingWorker* worker = new SmoothingWorker(this);
    worker->moveToThread(smoothThread);
    connect(smoothThread, SIGNAL(started()), worker, SLOT(process()));
    smoothThread->start();
}

void IGLMesh::loadScaleConstraints(QString filename, double scoreThreshold)
{
    QFile file(filename);
    file.open(QIODevice::ReadOnly);
    QTextStream in(&file);
    while (!in.atEnd())
    {
        QString line = in.readLine();
        QStringList lstLine = line.split(" ");

        int ktop = lstLine[15].toInt();
        double stop = lstLine[16].toDouble();
        double ttop = lstLine[17].toDouble();

        int kbot = lstLine[18].toInt();
        double sbot = lstLine[19].toDouble();
        double tbot = lstLine[20].toDouble();

        double score = lstLine[11].toDouble();
        if(score >= scoreThreshold)
        {
            floatingScaleConstraint.clear();
            floatingScaleConstraint.setP0Fixed(kbot,sbot,tbot);
            floatingScaleConstraint.setP1Fixed(ktop,stop,ttop);
            finishScaleConstraint(false);
        }
    }
    allowSmoothing();
}

void IGLMesh::getIncidentFaces()
{
    for(int i = 0; i < V.cols(); ++i)
    {
        incidentFaces.push_back(QVector<int>());
    }

    for(int j = 0; j < F.cols(); ++j)
    {
        int v0 = F(0,j);
        int v1 = F(1,j);
        int v2 = F(2,j);

        incidentFaces[v0].push_back(j);
        incidentFaces[v1].push_back(j);
        incidentFaces[v2].push_back(j);

        perFaceScaleConstraintList.push_back(QVector<int>());
        perFaceVertexConstraintList.push_back(QVector< QPair<int,int> >());
    }
}

void IGLMesh::centerMesh()
{
    for(int i = 0; i < V.cols(); ++i)
    {
        Vector3d vi = V.col(i);
        vi = vi - C;
        V.col(i) = vi;
    }
    C = Vector3d(0,0,0);
}

void IGLMesh::orientMesh()
{
    Vector3d Z(0,0,1);
    Vector3d axis = Z.cross(aveNormal);
    axis.normalize();

    double costheta = Z.dot(aveNormal);
    double angle = qAcos(costheta);


    double *quat = new double[4];
    double *rotmat = new double[16];
    double axis2[3] = {axis[0],axis[1],axis[2]};
    igl::axis_angle_to_quat(&axis2[0],angle,quat);
    igl::quat_to_mat(quat, rotmat);

    Matrix3d R;
    R << rotmat[0], rotmat[1], rotmat[2], rotmat[4], rotmat[5], rotmat[6], rotmat[8], rotmat[9], rotmat[10];

    for(int i = 0; i < V.cols(); ++i)
    {
        Vector3d vi = V.col(i);
        vi = R * vi;
        V.col(i) = vi;
    }

    delete[] quat;
    delete[] rotmat;

    computeNormals();
}

void IGLMesh::computeNormals()
{
    TN.resize(3,F.cols());
    VN.resize(3,V.cols());
    VN.setZero(3,V.cols());

    aveNormal = Vector3d(0,0,0);

    for(int i = 0; i < F.cols(); ++i)
    {
        Vector3d v0 = V.col(F(0,i));
        Vector3d v1 = V.col(F(1,i));
        Vector3d v2 = V.col(F(2,i));

        Vector3d v01 = v1 - v0;
        Vector3d v02 = v2 - v0;

        Vector3d ni = v01.cross(v02);
        ni.normalize();
        TN.col(i) = ni;
        aveNormal = aveNormal + ni;

        VN.col(F(0,i)) = VN.col(F(0,i)) + ni;
        VN.col(F(1,i)) = VN.col(F(1,i)) + ni;
        VN.col(F(2,i)) = VN.col(F(2,i)) + ni;
    }

    aveNormal = aveNormal / F.cols();
    aveNormal.normalize();

    for(int i = 0; i < V.cols(); ++i)
    {
        VN.col(i) = VN.col(i) / VN.col(i).norm();
    }
}

bool IGLMesh::loadMaterials(QString objFilename, QString mtlFilename)
{
    // OBJ FILE
    QFile objfile(objFilename);
	if(!objfile.open(QIODevice::ReadOnly)) 
    {
		qDebug() << "Can't find " << objFilename << "\n";
	}
	QTextStream in(&objfile);

    int currentFace = 0;
	int currentMat = 0;
    Materials.setZero(F.cols());
	while ( !in.atEnd() )
	{
		QString line = in.readLine();    
		QStringList fields = line.split(" ");
		if(fields.size() >= 1)
		{
			if(fields[0] == "usemtl")
			{
                QString matdata = fields[1];
                QStringList fields2 = line.split("_");
                QString mc = fields2[1];
                currentMat = mc.toInt();
            }
            else if(fields[0] == "f")
            {
                Materials(currentFace) = currentMat;
                ++currentFace;
            }
            if(currentFace >= F.cols())
                break;
        }
    }
    objfile.close();


    // MTL FILE
    QFile mtlfile(mtlFilename);
	if(!mtlfile.open(QIODevice::ReadOnly)) 
    {
        mtlFilename.chop(8);
        mtlFilename.append(".mtl");
        mtlfile.setFileName(mtlFilename);
        if(!mtlfile.open(QIODevice::ReadOnly))
        {
            qDebug() << "Can't find " << mtlFilename << "\n";
        }
	}
	QTextStream mtlin(&mtlfile);

    texFNames.clear();
	while ( !mtlin.atEnd() )
	{
		QString line = mtlin.readLine();    
		QStringList fields = line.split(" ");
		if(fields.size() >= 1)
		{
			if(fields[0] == "map_Kd")
			{
                QString texname = fields[1];
                texFNames.push_back(texname);
            }
        }
    }
    mtlfile.close();
    return true;
}

Vector3d IGLMesh::getSurfaceNormal(int k, double u, double v)
{
    Vector3d n0 = VN.col(F(0,k));
    Vector3d n1 = VN.col(F(1,k));
    Vector3d n2 = VN.col(F(2,k));

    Vector3d nOut = (1-u-v)*n0 + u*n1 + v*n2;
    nOut.normalize();
    return nOut;
}

void IGLMesh::setVertexPositions(MatrixXd newV)
{
    V = newV;
    C = Vector3d(0,0,0);
    for(int k = 0; k < V.cols(); ++k)
	{
        Vector3d R = V.col(k);
		C = C + R;
	}
    C = C / V.cols();
}

void IGLMesh::setInitialVertexPositions(MatrixXd initV)
{
    Vinitial = initV;
}

void IGLMesh::setFlatVertexPositions(MatrixXd newV)
{
    flatV = newV;
    flatC = Vector3d(0,0,0);
    for(int k = 0; k < flatV.cols(); ++k)
    {
        Vector3d R = flatV.col(k);
        flatC = flatC + R;
    }
    flatC = flatC / flatV.cols();

    for(int i = 0; i < flatV.cols(); ++i)
    {
        Vector3d vi = flatV.col(i);
        vi = vi - flatC;
        flatV.col(i) = vi;
    }

    flatC = Vector3d(0,0,0);
}

void IGLMesh::gatherVectors(int f_id)
{
    QVector<int> f_vertex_ids;
    f_vertex_ids.push_back(F(0,f_id));
    f_vertex_ids.push_back(F(1,f_id));
    f_vertex_ids.push_back(F(2,f_id));
    bool anythingGathered = false;

    anythingGathered = false;
    Vector3d v_right_F = Vector3d(0,0,0);
    for(int s = 0; s < perFaceVertexConstraintList[f_id].size(); ++s)
    {
        int l_index = perFaceVertexConstraintList[f_id][s].first;
        int lp_index = perFaceVertexConstraintList[f_id][s].second;
        LineConstraint lc = lineConstraints[l_index];
        int lpk = lc.new_faces[lp_index];
        if(lc.isActive())
        {
            Vector3d lp_right = lc.rightVectorsW.col(lp_index);
            if(lpk == f_id)
            {
                v_right_F = v_right_F + lp_right;
                anythingGathered = true;
            }
        }
    }
    if(anythingGathered)
    {
        v_right_F.normalize();
        isFaceVectorConstrained(f_id) = 1;
        perFaceRightVectorsW.col(f_id) = v_right_F;
    }
    else
    {
        isFaceVectorConstrained(f_id) = 0;
    }
}

void IGLMesh::reGatherCurrentConstraints()
{
    disallowSmoothing();
    if(lineConstraints.size() > 0)
    {
        for(int i = 0; i < lineConstraints.size(); ++i)
        {
            computeSingleLineConstraintRightAndUpVectors(&lineConstraints[i]);
            for(int lcp = 0; lcp < lineConstraints[i].size(); ++lcp)
            {
                int lcp_face = lineConstraints[i].faces[lcp];
                gatherVectors(lcp_face);
                //gatherVectorsFlat(lcp_face);
            }
        }
    }

    if(scaleConstraints2.size() > 0)
    {
        for(int i = 0; i < scaleConstraints2.size(); ++i)
        {
            int sc_face = scaleConstraints2[i].face;
            gatherScales(sc_face);
        }
    }

    allowSmoothing();
}

void IGLMesh::gatherScales(int f_id)
{
    QVector<int> f_vertex_ids;
    f_vertex_ids.push_back(F(0,f_id));
    f_vertex_ids.push_back(F(1,f_id));
    f_vertex_ids.push_back(F(2,f_id));

    for(int i = 0; i < 3; ++i)
    {
        int v_index = f_vertex_ids[i];
        Vector3d v_point = V.col(v_index);
        Vector2d scale_i(0,0);
        double total_weight_i = 0;
        bool anythingGathered = false;
        for(int j = 0; j < incidentFaces[v_index].size(); ++j)
        {
            int f_index = incidentFaces[v_index][j];

            for(int s = 0; s < perFaceScaleConstraintList[f_index].size(); ++s)
            {
                int s_index = perFaceScaleConstraintList[f_index][s];
                ScaleConstraint sc = scaleConstraints2[s_index];
                if(sc.isActive())
                {
                    int sk = sc.face;
                    double su = sc.u;
                    double sv = sc.v;
                    Vector3d sc_point = getWrinkledSurfacePoint(sk,su,sv);
                    double sc_dist = (sc_point - v_point).norm();
                    Vector2d sc_scale = sc.scale;

                    if(sk == f_index)
                    {
                        double w = (1.0 / sc_dist);
                        scale_i = scale_i + w * sc_scale;
                        total_weight_i += w;
                        anythingGathered = true;
                    }
                }
            }
        }
        if(anythingGathered)
        {
            scale_i = scale_i / total_weight_i;
            isScaleConstrained(v_index) = 1;
            perVertexScaleW.col(v_index) = scale_i;
        }
        else
        {
            isScaleConstrained(v_index) = 0;
        }
    }
}

void IGLMesh::selectedLineConstraintForward()
{
    selectedLineConstraint = selectedLineConstraint + 1;
    if(selectedLineConstraint == lineConstraints.size())
    {
        selectedLineConstraint = -1;
        ParchmentFlattener::markupMode = ParchmentFlattener::Line;
    }
    else
    {
        ParchmentFlattener::markupMode = ParchmentFlattener::EditLine;
        if(lineConstraints[selectedLineConstraint].isActive() == false)
        {
            selectedLineConstraintForward();
        }
    }
}

void IGLMesh::selectedScaleConstraintForward()
{
    selectedScaleConstraint = selectedScaleConstraint + 1;
    if(selectedScaleConstraint == scaleConstraints2.size())
    {
        selectedScaleConstraint = -1;
        ParchmentFlattener::markupMode = ParchmentFlattener::Scale;
    }
    else
    {
        ParchmentFlattener::markupMode = ParchmentFlattener::EditScale;
        if(scaleConstraints2[selectedScaleConstraint].isActive() == false)
        {
            selectedScaleConstraintForward();
        }
    }
}

void IGLMesh::selectedLineConstraintBackward()
{
    if(selectedLineConstraint >= 0)
        selectedLineConstraint = selectedLineConstraint - 1;
    else
        selectedLineConstraint = lineConstraints.size() - 1;
    if(selectedLineConstraint == -1)
    {
        ParchmentFlattener::markupMode = ParchmentFlattener::Line;
    }
    else
    {
        ParchmentFlattener::markupMode = ParchmentFlattener::EditLine;
        if(lineConstraints[selectedLineConstraint].isActive() == false)
        {
            selectedLineConstraintBackward();
        }
    }
}

void IGLMesh::selectedScaleConstraintBackward()
{
    if(selectedScaleConstraint >= 0)
        selectedScaleConstraint = selectedScaleConstraint - 1;
    else
        selectedScaleConstraint = scaleConstraints2.size() - 1;
    if(selectedScaleConstraint == -1)
    {
        ParchmentFlattener::markupMode = ParchmentFlattener::Scale;
    }
    else
    {
        ParchmentFlattener::markupMode = ParchmentFlattener::EditScale;
        if(scaleConstraints2[selectedScaleConstraint].isActive() == false)
        {
            selectedScaleConstraintBackward();
        }
    }
}

Vector3d IGLMesh::getWrinkledSurfacePoint(int k, double s, double t)
{
    int vi0 = F(0,k);
    int vi1 = F(1,k);
    int vi2 = F(2,k);

    Vector3d v0 = V.col(vi0);
    Vector3d v1 = V.col(vi1);
    Vector3d v2 = V.col(vi2);

    Vector3d vOut = (1 - s - t) * v0 + s * v1 + t * v2;
    return vOut;
}

Vector3d IGLMesh::getFlatSurfacePoint(int k, double s, double t)
{
    int vi0 = F(0,k);
    int vi1 = F(1,k);
    int vi2 = F(2,k);

    Vector3d v0 = flatV.col(vi0);
    Vector3d v1 = flatV.col(vi1);
    Vector3d v2 = flatV.col(vi2);

    Vector3d vOut = (1 - s - t) * v0 + s * v1 + t * v2;
    return vOut;
}

Vector3d IGLMesh::getInitSurfacePoint(int k, double s, double t)
{
    int vi0 = F(0,k);
    int vi1 = F(1,k);
    int vi2 = F(2,k);

    Vector3d v0 = Vinitial.col(vi0);
    Vector3d v1 = Vinitial.col(vi1);
    Vector3d v2 = Vinitial.col(vi2);

    Vector3d vOut = (1 - s - t) * v0 + s * v1 + t * v2;
    return vOut;
}

void IGLMesh::clickLineConstraint(int k, double s, double t)
{
    disallowSmoothing();
    floatingLineConstraint.addPoint(k,s,t);
    floatingLineConstraint.clearSuggestions();
}

void IGLMesh::finishLineConstraint(bool isVertical)
{
    if(floatingLineConstraint.size() > 1)
    {
        floatingLineConstraint.mIsVertical = isVertical;
        computeSingleLineConstraintRightAndUpVectors(&floatingLineConstraint);
        lineConstraints.push_back(floatingLineConstraint);

    }
    floatingLineConstraint.clear();

    int lc_id = lineConstraints.size() - 1;
    for(int lcp = 0; lcp < lineConstraints[lc_id].size(); ++lcp)
    {
        int lcp_face = lineConstraints[lc_id].faces[lcp];
        perFaceVertexConstraintList[lcp_face].push_back(QPair<int,int>(lc_id,lcp));
        gatherVectors(lcp_face);
    }

    needsFlattening = true;

    emit autoSave();

    allowSmoothing();
}

void IGLMesh::finishEditingLineConstraint()
{
    if(aLineConstraintWasMoved)
    {
        computeSingleLineConstraintRightAndUpVectors(&(lineConstraints[selectedLineConstraint]));

        int oldface = lineConstraints[selectedLineConstraint].faces[selectedLineConstraintPoint];
        int newface = lineConstraints[selectedLineConstraint].new_faces[selectedLineConstraintPoint];

        perFaceVertexConstraintList[oldface].remove(perFaceVertexConstraintList[oldface].indexOf(QPair<int,int>(selectedLineConstraint, selectedLineConstraintPoint)));
        perFaceVertexConstraintList[newface].push_back(QPair<int,int>(selectedLineConstraint, selectedLineConstraintPoint));

        gatherVectors(oldface);
        for(int k = 0; k < lineConstraints[selectedLineConstraint].size(); ++k)
        {
            int lcpface = lineConstraints[selectedLineConstraint].new_faces[k];
            gatherVectors(lcpface);

            lineConstraints[selectedLineConstraint].faces[k] = lineConstraints[selectedLineConstraint].new_faces[k];
            lineConstraints[selectedLineConstraint].sts(0,k) = lineConstraints[selectedLineConstraint].new_sts(0,k);
            lineConstraints[selectedLineConstraint].sts(1,k) = lineConstraints[selectedLineConstraint].new_sts(1,k);
        }
        aLineConstraintWasMoved = false;
        needsFlattening = true;

        emit autoSave();

        allowSmoothing();
    }
}

void IGLMesh::deleteCurrentLineConstraint()
{
    disallowSmoothing();
    while(!smoothIsPaused)
    {
        qDebug() << "wating for current smoothing loop to finish";
    }
    lineConstraints[selectedLineConstraint].deleteLineConstraint();
    for(int k = 0; k < lineConstraints[selectedLineConstraint].size(); ++k)
    {
        int lcpface = lineConstraints[selectedLineConstraint].faces[k];
        gatherVectors(lcpface);
    }
    selectedLineConstraint = -1;
    needsFlattening = true;

    emit autoSave();

    allowSmoothing();
}

void IGLMesh::deleteCurrentScaleConstraint()
{
    disallowSmoothing();
    while(!smoothIsPaused)
    {
        qDebug() << "wating for current smoothing loop to finish";
    }
    scaleConstraints2[selectedScaleConstraint].deleteScaleConstraint();
    int scface = scaleConstraints2[selectedScaleConstraint].face;
    gatherScales(scface);
    selectedScaleConstraintForward();
    needsFlattening = true;

    emit autoSave();

    allowSmoothing();
}

void IGLMesh::clearConstraints()
{
    disallowSmoothing();
    lineConstraints.clear();
    scaleConstraints2.clear();
    reGatherCurrentConstraints();
    needsFlattening = true;

    emit autoSave();

    allowSmoothing();
}

void IGLMesh::moveSelectedLineConstraintPoint(int k, double s, double t)
{
    if(selectedLineConstraint >= 0 && selectedLineConstraint < lineConstraints.size())
    {
        if(selectedLineConstraintPoint >= 0 && selectedLineConstraintPoint < lineConstraints[selectedLineConstraint].size())
        {
            lineConstraints[selectedLineConstraint].new_faces[selectedLineConstraintPoint] = k;
            lineConstraints[selectedLineConstraint].new_sts(0,selectedLineConstraintPoint) = s;
            lineConstraints[selectedLineConstraint].new_sts(1,selectedLineConstraintPoint) = t;
            aLineConstraintWasMoved = true;
        }
    }
}

void IGLMesh::computeSingleLineConstraintRightAndUpVectors(LineConstraint *lc)
{
    lc->rightVectorsW.resize(3,lc->size());
    lc->upVectorsW.resize(3,lc->size());

    lc->rightVectorsF.resize(3,lc->size());
    lc->upVectorsF.resize(3,lc->size());

    for(int i = 0; i < lc->size(); ++i)
    {
        int ki = lc->new_faces[i];
        double ui = lc->new_sts(0,i);
        double vi = lc->new_sts(1,i);

        int kleft, kright;
        double uleft, uright;
        double vleft, vright;

        Vector3d leftNeighbourW;
        Vector3d rightNeighbourW;
        Vector3d leftNeighbourF;
        Vector3d rightNeighbourF;

        if(i == 0)
        {
            kleft = ki;
            uleft = ui;
            vleft = vi;
            kright = lc->new_faces[i+1];
            uright = lc->new_sts(0,i+1);
            vright = lc->new_sts(1,i+1);
        }
        else if(i == lc->size() - 1)
        {
            kleft = lc->new_faces[i-1];
            uleft = lc->new_sts(0,i-1);
            vleft = lc->new_sts(1,i-1);
            kright = ki;
            uright = ui;
            vright = vi;
        }
        else
        {
            kleft = lc->new_faces[i-1];
            uleft = lc->new_sts(0,i-1);
            vleft = lc->new_sts(1,i-1);
            kright = lc->new_faces[i+1];
            uright = lc->new_sts(0,i+1);
            vright = lc->new_sts(1,i+1);
        }

        leftNeighbourW = getWrinkledSurfacePoint(kleft,uleft,vleft);
        rightNeighbourW = getWrinkledSurfacePoint(kright,uright,vright);

        leftNeighbourF = getFlatSurfacePoint(kleft,uleft,vleft);
        rightNeighbourF = getFlatSurfacePoint(kright,uright,vright);

        Vector3d rightVecW = rightNeighbourW - leftNeighbourW;
        rightVecW.normalize();

        Vector3d rightVecF = rightNeighbourF - leftNeighbourF;
        rightVecF.normalize();

        Vector3d ni = getSurfaceNormal(ki,ui,vi);
        Vector3d upVecW = ni.cross(rightVecW);

        Vector3d niF = Vector3d(0,0,1);
        Vector3d upVecF = niF.cross(rightVecF);

        if(lc->mIsVertical)
        {
            Vector3d tempVec;

            tempVec = rightVecW;
            rightVecW = upVecW;
            upVecW = -tempVec;

            tempVec = rightVecF;
            rightVecF = upVecF;
            upVecF = -tempVec;
        }

        lc->rightVectorsW.col(i) = rightVecW;
        lc->upVectorsW.col(i) = upVecW;

        lc->rightVectorsF.col(i) = rightVecF;
        lc->upVectorsF.col(i) = upVecF;
    }
}

void IGLMesh::computeAllLineConstraintRightAndUpVectors()
{
    for(int i = 0; i < lineConstraints.size(); ++i)
    {
        computeSingleLineConstraintRightAndUpVectors(&lineConstraints[i]);
    }
}

double IGLMesh::getDiffusionWeightBetweenVertices(int v0, int v1)
{
    Vector3d n0 = VN.col(v0);
    Vector3d n1 = VN.col(v1);
    double weight = n0.dot(n1);
    if(weight < 0)
        weight = 0;
    weight = qPow(weight, 4);
    return weight;
}

// Solves a linear system Ax=c with given boundary conditions defined in boundaryConditions as indicies of the known rows
template<typename T>
static void SolveLinearSystemWithBoundaryConditions(const Eigen::SparseMatrix<T>& A, Eigen::Matrix<T,Eigen::Dynamic,1>& x, const Eigen::Matrix<T,Eigen::Dynamic,1>& c, std::vector<int>& boundaryConditions)
{
    // sort boundary conditions indices
    std::sort(boundaryConditions.begin(),boundaryConditions.end());

    const int size = x.size();
    Eigen::SparseMatrix<T> AP(size,size);

    const int knownCount = boundaryConditions.size();
    const int unknownCount = size - knownCount;
    unsigned int knownIndex = 0;
    unsigned int unknownIndex = 0;

    // Permutate to sort by known boundary conditions -> [A1,A2] * [x1;x2]
    Eigen::VectorXi transpositions;
    transpositions.resize(size);
    for(int n=0;n<size;n++)
    {
      if(knownIndex < boundaryConditions.size() && n == boundaryConditions[knownIndex])
      {
        transpositions[unknownCount + knownIndex] = n;
        knownIndex++;
      }
      else
      {
        transpositions[unknownIndex] = n;
        unknownIndex++;
      }
    }
    PermutationMatrix<Dynamic,Dynamic,int> P(transpositions);
    P = P.inverse();
    A.twistedBy(P).evalTo(AP);

    // Split in kown and unknown parts A1,A2,x2,c1
    SparseMatrix<T> A1, A2;
    Matrix<T,Dynamic,1> x1, x2, c1;
    x2 = (P*x).block(unknownCount,0,knownCount,1);
    c1 = c.block(0,0,unknownCount,1);

    SparseMatrix<T> A1Temp(size,unknownCount);
    SparseMatrix<T> A2Temp(size,knownCount);
    A1Temp = AP.innerVectors(0,unknownCount).transpose();
    A2Temp = AP.innerVectors(unknownCount,knownCount).transpose();
    A1 = A1Temp.innerVectors(0,unknownCount).transpose();
    A2 = A2Temp.innerVectors(0,unknownCount).transpose();

    // Solve A1*x1 = c - A2*x2
    //SparseQR<SparseMatrix<T>, COLAMDOrdering<int> > solver;
    //qDebug() << "solver.compute(A1)";
    //solver.compute(A1);
    //qDebug() << "finished solver.compute(A1)";



    SparseLU<SparseMatrix<T>, COLAMDOrdering<int> > solver;
    solver.analyzePattern(A1);
    solver.factorize(A1);

    if (solver.info() != Eigen::Success)
    {
      std::cerr << "Matrix can not be decomposed.\n";
      return;
    }

    x1 = solver.solve(c1-A2*x2);

    if (solver.info() != Eigen::Success)
    {
      std::cerr << "Solving failed.\n";
      return;
    }

    // Invert permutation and create resulting x
    x.block(0,0,unknownCount,1) = x1;
    x.block(unknownCount,0,knownCount,1) = x2;
    x = P.inverse() * x;
}

void ComputeScaleField(const Eigen::MatrixX3d& V, const Eigen::MatrixX3i& F, const Eigen::MatrixX3d& N, std::vector<int> contraintIdx, Eigen::VectorXd& x, bool withCot)
{
  // sparse weight matrix W
  Eigen::SparseMatrix<double> W;

  //compute face-face adjacency matrix
  Eigen::MatrixX3i TTi;
  igl::tt(V,F,TTi);

  qDebug() << "Init Weight Matrix";
  // create weight matrix W
  W.resize(V.rows(),V.rows());
  std::vector<Eigen::Triplet<double> > triplets;
  for(int fA=0;fA<F.rows();fA++)
  {
    for(int i=0;i<3;i++)
    {
        // corresponding edge
        int v1 = F(fA,i);
        int v2 = F(fA,(i+1)%3);
        Vector3d nA = N.row(v1);
        Vector3d nB = N.row(v2);
        // compute alignment weight
        double v = 1.0;//(nA.dot(nB)+1)/2;
        v = pow(v,4);

        // add weight to W
        triplets.push_back(Eigen::Triplet<double>(v1,v2,v));
        // check if boundary
        if(TTi(fA,i) == -1)
          triplets.push_back(Eigen::Triplet<double>(v2,v1,v));
    }
  }

  // initialize W
  for(int i=0;i<W.cols();i++)
      triplets.push_back(Eigen::Triplet<double>(i,i,1));
  W.setFromTriplets(triplets.begin(), triplets.end());

  qDebug() << "Multiply by Laplacian";
  // multiply by Laplacian
  if(withCot)
  {
    // create sparse laplacian matrix L
    Eigen::SparseMatrix<double> L;
    igl::cotmatrix(V,F,L);
    L *= -1;

    // multiply L by weight matrix coeffwise
    Eigen::SparseMatrix<double> L2;
    W = W.cwiseProduct(L);
  }

  qDebug() << "Normalize Columns";
  // normalize columns of W
  std::vector<double> colSums;
  std::vector<double*> WDiag;
  for(int i=0;i<W.cols();i++)
    colSums.push_back(0);

  qDebug() << "Sum Columns";
  // sum up columns
  for (int k=0;k<W.outerSize();++k)
  {
    for (Eigen::SparseMatrix<double>::InnerIterator it(W,k);it;++it)
    {
      if(it.col() == it.row())
        WDiag.push_back(&it.valueRef());
      else
        colSums[it.col()] += it.value();
    }
  }

  qDebug() << "Set Diag";
  // set diagonal of W
  for(int i=0;i<W.cols();i++)
    *WDiag[i] = -colSums[i];

  qDebug() << "Solve";
  // Solver Lx=0 s.t. given constraints
  Eigen::VectorXd b = VectorXd::Zero(x.rows());
  //L2 = L2*L2;
  SolveLinearSystemWithBoundaryConditions(W,x,b,contraintIdx);
}

void IGLMesh::computeVectorAndScalesDiffusion()
{
    // NRosy for field
    MatrixX3d Vt = V.transpose();
    MatrixX3i Ft = F.transpose();

    MatrixX3d VNt = VN.transpose();
    std::vector<int> constrainedScales;
    VectorXd scaleX = perVertexScaleW.row(0);
    VectorXd scaleY = perVertexScaleW.row(1);
    for(int i = 0; i < V.cols(); ++i)
    {
        if(isScaleConstrained(i) == 1)
        {
            constrainedScales.push_back(i);
            scaleX(i) = perVertexScaleW(0,i);
            scaleY(i) = perVertexScaleW(1,i);
        }
        else
        {
            scaleX(i) = 0.0;
            scaleY(i) = 0.0;
        }
    }

    if(constrainedScales.size() > 0)
    {
        ComputeScaleField(Vt,Ft,VNt,constrainedScales,scaleX,true);
        ComputeScaleField(Vt,Ft,VNt,constrainedScales,scaleY,true);
        perVertexScaleW.row(0) = scaleX;
        perVertexScaleW.row(1) = scaleY;

        /*
        for(int i = 0; i < scaleX.rows(); ++i)
        {
            if(scaleX(i) < 0)
                qDebug() << "negative at " << i;
        }
        */
    }

    NRosyField *nrosy = new NRosyField(Vt,Ft);
    for(int k = 0; k < F.cols(); ++k)
    {
        if(isFaceVectorConstrained(k) == 1)
        {
            Vector3d vright_k = perFaceRightVectorsW.col(k);
            nrosy->setConstraintHard(k,vright_k);
        }
    }
    qDebug() << "Solving NRosy";
    nrosy->solve(1);
    MatrixXd nrosyfield = nrosy->getFieldPerFace();
    perFaceRightVectorsW = nrosyfield.transpose();
    delete nrosy;
    qDebug() << "Solved NRosy " << nrosyfield.rows() << " x " << nrosyfield.cols();

    for(int k = 0; k < F.cols(); ++k)
    {
        Vector3d nk = TN.col(k);
        Vector3d rk = perFaceRightVectorsW.col(k);
        Vector3d uk = nk.cross(rk);
        uk.normalize();
        perFaceUpVectorsW.col(k) = uk;
    }
}

void IGLMesh::transferVertexGradientToFaces3D()
{
    qDebug() << "Transferring vertex crossfield to faces";
    perFaceRightVectorsW.resize(3,F.cols());
    perFaceUpVectorsW.resize(3,F.cols());
    perFaceScaleW.setOnes(2,F.cols());
    for(int k = 0; k < F.cols(); ++k)
    {
        int v0 = F(0,k);
        int v1 = F(1,k);
        int v2 = F(2,k);

        Vector2d sc0 = perVertexScaleW.col(v0);
        Vector2d sc1 = perVertexScaleW.col(v1);
        Vector2d sc2 = perVertexScaleW.col(v2);
        Vector2d scalek = (sc0 + sc1 + sc2) / 3.0;

        perFaceScaleW.col(k) = scalek;
    }
}

void IGLMesh::transferCrossFieldFrom3Dto2D()
{
    qDebug() << "Transferring crossfield to 2D";

    perFaceRightVectorsF.setZero(3,F.cols());
    perFaceUpVectorsF.setZero(3,F.cols());
    perFaceScaleF.setZero(2,F.cols());

    aveRightVectorF = Vector3d(0,0,0);

    for(int k = 0; k < F.cols(); ++k)
    {
        int v0 = F(0,k);
        int v1 = F(1,k);
        int v2 = F(2,k);
        Vector3d refEdge3D = V.col(v1) - V.col(v0);

        refEdge3D.normalize();
        Vector3d rk = perFaceRightVectorsW.col(k);
        Vector3d uk = perFaceUpVectorsW.col(k);
        Vector3d nk = TN.col(k);

        double theta_r = qAcos(rk.dot(refEdge3D));
        double theta_u = qAcos(uk.dot(refEdge3D));

        Vector3d cross_r = refEdge3D.cross(rk);
        Vector3d cross_u = refEdge3D.cross(uk);
        cross_r.normalize();
        cross_u.normalize();
        if (nk.dot(cross_r) < 0)
            theta_r = -theta_r;
        if (nk.dot(cross_u) < 0)
            theta_u = -theta_u;

        Vector3d refEdge2D = flatV.col(v1) - flatV.col(v0);
        refEdge2D.normalize();

        double csr = qCos(theta_r);
        double snr = qSin(theta_r);
        double rx = refEdge2D(0) * csr + refEdge2D(1) * -snr;
        double ry = refEdge2D(0) * snr + refEdge2D(1) * csr;

        Vector3d rf(rx,ry,0);
        //rf.normalize();
        perFaceRightVectorsF.col(k) = rf;
        Vector3d uf = Vector3d(0,0,1).cross(rf);
        uf.normalize();
        perFaceUpVectorsF.col(k) = uf;

        if(rx == rx && ry == ry)
        {
            aveRightVectorF = aveRightVectorF + Vector3d(rx,ry,0);
        }
        else
        {
            qDebug() << "NaN right vector";
        }

        // Scale
        Vector3d V02 = V.col(v2) - V.col(v0);
        Vector3d V01 = V.col(v1) - V.col(v0);
        Vector3d FV02 = flatV.col(v2) - flatV.col(v0);
        Vector3d FV01 = flatV.col(v1) - flatV.col(v0);
        Vector3d crossW = V01.cross(V02);
        Vector3d crossF = FV01.cross(FV02);
        double areaW = crossW.norm();
        double areaF = crossF.norm();
        double FoverW = qSqrt(areaF / areaW);

        Vector2d scaleW = perFaceScaleW.col(k);
        Vector2d scaleF = scaleW * FoverW;
        //perFaceScaleF.col(k) = scaleF;
        perFaceScaleF.col(k) = scaleW;
    }

    aveRightVectorF.normalize();
    Vector3d aveUpVec = Vector3d(0,0,1).cross(aveRightVectorF);
    aveUpVec.normalize();

    for(int k = 0; k < F.cols(); ++k)
    {
        Vector3d rk = perFaceRightVectorsF.col(k);
        if(rk.dot(aveRightVectorF) < 0)
        {
            perFaceRightVectorsF.col(k) =aveRightVectorF;
            perFaceUpVectorsF.col(k) = aveUpVec;
        }
    }
}

void IGLMesh::smooth2DCrossField()
{
    MatrixXd newPerFaceRightVectorsF;
    newPerFaceRightVectorsF.setZero(perFaceRightVectorsF.rows(), perFaceRightVectorsF.cols());

    aveRightVectorF = Vector3d(0,0,0);
    for(int k = 0; k < F.cols(); ++k)
    {
        int v0 = F(0,k);
        int v1 = F(1,k);
        int v2 = F(2,k);

        Vector3d rk = perFaceRightVectorsF.col(k);

        QSet<int> faces;
        for(int i = 0; i < incidentFaces[v0].size(); ++i)
            faces.insert(incidentFaces[v0][i]);
        for(int i = 0; i < incidentFaces[v1].size(); ++i)
            faces.insert(incidentFaces[v1][i]);
        for(int i = 0; i < incidentFaces[v2].size(); ++i)
            faces.insert(incidentFaces[v2][i]);
        faces.remove(k);
        QVector<int> facesV = faces.toList().toVector();

        Vector3d rn(0,0,0);
        for(int j = 0; j < facesV.size(); ++j)
        {
            int face_index = facesV[j];
            Vector3d rj = perFaceRightVectorsF.col(face_index);
            rn = rn + rj;
        }
        rn.normalize();

        Vector3d rk_new = rk*0.5 + rn*0.5;
        rk_new.normalize();
        Vector3d uk_new = Vector3d(0,0,1).cross(rk_new);
        //uk_new = uk_new + 5 * rk_new;
        uk_new.normalize();
        newPerFaceRightVectorsF.col(k) = rk_new;
        perFaceUpVectorsF.col(k) = uk_new;

        aveRightVectorF = aveRightVectorF + rk_new;
    }
    aveRightVectorF.normalize();
    perFaceRightVectorsF = newPerFaceRightVectorsF;
}

void IGLMesh::alignAveRightVecWithXAxis()
{
    Vector3d xaxis(1,0,0);
    Matrix3d R;
    Quaterniond Q;
    Q.setFromTwoVectors(aveRightVectorF,xaxis);
    R = Q.toRotationMatrix();
    qDebug() << R(0,0) << " " << R(0,1) << " " << R(0,2);
    qDebug() << R(1,0) << " " << R(1,1) << " " << R(1,2);
    qDebug() << R(2,0) << " " << R(2,1) << " " << R(2,2);
    for(int i = 0; i < flatV.cols(); ++i)
    {
        Vector3d vi = flatV.col(i);
        Vector3d vi2 = R * vi;
        flatV.col(i) = vi2;

        vi = Vinitial.col(i);
        vi2 = R* vi;
        Vinitial.col(i) = vi2;
    }
    isAligned = true;
}

void IGLMesh::workerFinished()
{
    qDebug() << "WORKER FINISHED";
}

void IGLMesh::doAllSmoothing()
{
    //bool finished = false;
    int numIter = 0;
    while(true)
    {
        //qDebug() << smoothingAllowed;
        sleep(1);
        if(smoothingAllowed)
        {
            smoothIsPaused = false;
            qDebug() << "Doing Smoothing";
            computeVectorAndScalesDiffusion();
            transferVertexGradientToFaces3D();
            transferCrossFieldFrom3Dto2D();
            if(!isAligned)
                alignAveRightVecWithXAxis();
            transferCrossFieldFrom3Dto2D();
            smooth2DCrossField();
            emit smoothingIterationFinished();
            flatteningEngine->rebuildLimData(this);
            smoothingAllowed = false;
            smoothIsPaused = true;
            if(needsFlattening)
            {
                qDebug() << "Enabling Flattening";
                flatteningEnabled = true;
                numIter = 0;
            }
        }
        else if(!flatteningLocked)
        {
            smoothIsPaused = true;
            numIter = 0;
            while(flatteningEnabled)
            {
                qDebug() << "Doing Flattening";
                needsFlattening = false;
                int stepResult = flatteningEngine->singleStep(this);
                emit flatteningIterationFinished();
                if(stepResult > 0)
                    ++numIter;
                else
                    numIter = 0;
                if(numIter > 10)
                {
                    qDebug() << "Disable Flattening!";
                    flatteningEnabled = false;
                    aabbtree->rebuildTree();
                    transferCrossFieldFrom3Dto2D();
                }
            }
        }
    }
}

bool IGLMesh::intersectTriangleWithRay(Vector3d O, Vector3d D, Vector3d vert0, Vector3d vert1, Vector3d vert2, double *t, double *u, double *v)
{
    const double EPSILON = 0.000001;
    Vector3d edge1, edge2, tvec, pvec, qvec;
    double det, inv_det;
    edge1 = vert1 - vert0;
    edge2 = vert2 - vert0;
    pvec = D.cross(edge2);
    det = edge1.dot(pvec);
    if(det > -EPSILON && det < EPSILON)
        return false;
    inv_det = 1.0 / det;
    tvec = O - vert0;
    double _u = tvec.dot(pvec) * inv_det;
    if(_u < 0.0 || _u > 1.0)
        return false;
    qvec = tvec.cross(edge1);
    double _v = D.dot(qvec) * inv_det;
    if(_v < 0.0 || _v > 1.0)
        return false;
    double _w = 1 - _u - _v;
    if(_w < 0.0 || _w > 1.0)
        return false;
    *u = _u;
    *v = _v;
    *t = edge2.dot(qvec) * inv_det;
    return true;
}

bool IGLMesh::intersectMeshWithRay(Vector3d O, Vector3d D, double *t, int *k, double *u, double *v)
{
    double dist = 0;
    double best_dist = 9999999;
    int best_k = -1;
    bool any_intersect = false;

    for(int _k = 0; _k < F.cols(); ++_k)
    {
        int v0 = F(0,_k);
        int v1 = F(1,_k);
        int v2 = F(2,_k);

        Vector3d vert0 = Vuntouched.col(v0);
        Vector3d vert1 = Vuntouched.col(v1);
        Vector3d vert2 = Vuntouched.col(v2);

        double _u = 0;
        double _v = 0;

        bool is_intersect = intersectTriangleWithRay(O,D,vert0,vert1,vert2,&dist,&_u,&_v);
        double abs_dist = qAbs(dist);
        if(is_intersect && abs_dist < best_dist)
        {
            best_dist = abs_dist;
            best_k = _k;
            *u = _u;
            *v = _v;
            any_intersect = true;
        }
    }
    *k = best_k;
    *t = best_dist;
    return any_intersect;
}

bool IGLMesh::intersectMeshWithRayFlat(Vector3d O, Vector3d D, double *t, int *k, double *u, double *v)
{
    aabbtree->IntersectRay(O,D,t,k,u,v);
    if(*k >= 0)
        return true;
    else
    {
        double dist = 0;
        double best_dist = 9999999;
        int best_k = -1;
        bool any_intersect = false;

        for(int _k = 0; _k < F.cols(); ++_k)
        {
            int v0 = F(0,_k);
            int v1 = F(1,_k);
            int v2 = F(2,_k);

            Vector3d vert0 = flatV.col(v0);
            Vector3d vert1 = flatV.col(v1);
            Vector3d vert2 = flatV.col(v2);

            double _u = 0;
            double _v = 0;

            bool is_intersect = intersectTriangleWithRay(O,D,vert0,vert1,vert2,&dist,&_u,&_v);
            double abs_dist = qAbs(dist);
            if(is_intersect && abs_dist < best_dist)
            {
                best_dist = abs_dist;
                best_k = _k;
                *u = _u;
                *v = _v;
                any_intersect = true;
            }
        }
        *k = best_k;
        *t = best_dist;
        return any_intersect;
    }
}

LineConstraint::LineConstraint(void)
{
    sts = MatrixXd();
    new_sts = MatrixXd();
    rightVectorsW = MatrixXd();
    upVectorsW = MatrixXd();
    active = true;
    mIsVertical = false;
}

LineConstraint::~LineConstraint(void)
{
}

void LineConstraint::deleteLineConstraint()
{
    active = false;
    qDebug() << "LINE CONSTRAINT DEACTIVATED";
}

bool LineConstraint::isActive()
{
    return active;
}

void LineConstraint::addPoint(int k, double s, double t)
{
    faces.push_back(k);
    new_faces.push_back(k);
    int newNumpts = size() + 1;
    sts.conservativeResize(2,newNumpts);
    sts(0,newNumpts-1) = s;
    sts(1,newNumpts-1) = t;
    new_sts.conservativeResize(2,newNumpts);
    new_sts(0,newNumpts-1) = s;
    new_sts(1,newNumpts-1) = t;
}


int LineConstraint::size()
{
    return sts.cols();
}

void LineConstraint::clear()
{
    sts.resize(0,0);
    new_sts.resize(0,0);
    rightVectorsW.resize(0,0);
    upVectorsW.resize(0,0);
    faces.clear();
    new_faces.clear();
    mIsVertical = false;
}

void LineConstraint::addSuggestionPoint(int k, double s, double t)
{
    suggested_faces.push_back(k);
    int newNumpts = suggested_faces.size();
    suggested_sts.conservativeResize(2,newNumpts);
    suggested_sts(0,newNumpts-1) = s;
    suggested_sts(1,newNumpts-1) = t;
}

void LineConstraint::clearSuggestions()
{
    suggested_faces.clear();
    suggested_sts.resize(0,0);
}

void LineConstraint::acceptSuggestions()
{
    for(int i = 0; i < suggested_faces.size(); ++i)
    {
        int sk = suggested_faces[i];
        double ss = suggested_sts(0,i);
        double st = suggested_sts(1,i);
        addPoint(sk,ss,st);
    }
    clearSuggestions();
}

void LineConstraint::acceptSuggestions(int n)
{
    for(int i = 0; i < suggested_faces.size(); ++i)
    {
        if(i > n)
            break;
        int sk = suggested_faces[i];
        double ss = suggested_sts(0,i);
        double st = suggested_sts(1,i);
        addPoint(sk,ss,st);
    }
    clearSuggestions();
}

ScaleConstraint::ScaleConstraint(void)
{
    clear();
}

ScaleConstraint::~ScaleConstraint(void)
{
}

void ScaleConstraint::clear()
{
    scale = Vector2d(0,0);
    face = -1;
    u = -1;
    v = -1;
    face0 = -1;
    u0 = -1;
    v0 = -1;
    face1 = -1;
    u1 = -1;
    v1 = -1;
    face2 = -1;
    u2 = -1;
    v2 = -1;
    p0fixed = false;
    p1fixed = false;
    p2fixed = false;
    active = true;
}

void ScaleConstraint::setP0Floating(int k, double u, double v)
{
    if(!p0fixed)
    {
        face0 = k;
        u0 = u;
        v0 = v;
    }
}

void ScaleConstraint::setP0Fixed(int k, double u, double v)
{
    setP0Floating( k,  u,  v);
    p0fixed = true;
}

void ScaleConstraint::setP1Floating(int k, double u, double v)
{
    if(!p1fixed)
    {
        face1 = k;
        u1 = u;
        v1 = v;
    }
}

void ScaleConstraint::setP1Fixed(int k, double u, double v)
{
    setP1Floating( k,  u,  v);
    p1fixed = true;
}

void ScaleConstraint::setP2Floating(int k, double u, double v)
{
    if(!p2fixed)
    {
        face2 = k;
        u2 = u;
        v2 = v;
    }
}

void ScaleConstraint::setP2Fixed(int k, double u, double v)
{
    setP2Floating( k,  u,  v);
    p2fixed = true;
}

void ScaleConstraint::setMidPoint(int _face, double _u, double _v)
{
    face = _face;
    u = _u;
    v = _v;
}

bool ScaleConstraint::isActive()
{
    return active;
}

void ScaleConstraint::deleteScaleConstraint()
{
    active = false;
}

QDataStream &operator>>(QDataStream &in, ScaleConstraint &sc)
{
    QString letter_s;
    QVector3D pMid;
    QVector3D pTop;
    QVector3D pRight;
    double scale;
    double score;
    int kMid;
    double sMid, tMid;
    in >> letter_s >> pMid >> pTop >> pRight >> scale >> score >> kMid >> sMid >> tMid;
    sc = ScaleConstraint();
    sc.setMidPoint(kMid,sMid,tMid);
    return in;
}

void IGLMesh::clickScaleConstraint(int k, double s, double t, bool rightClick)
{
    disallowSmoothing();
    if(!rightClick)
    {
        if(!floatingScaleConstraint.p0fixed)
        {
            floatingScaleConstraint.setP0Fixed(k,s,t);
            floatingScaleConstraint.setP1Floating(k,s,t);
        }
        else
        {
            floatingScaleConstraint.setP1Fixed(k,s,t);
            finishScaleConstraint(true);
        }
    }
    else
    {
        if(!floatingScaleConstraint.p0fixed)
        {
            floatingScaleConstraint.setP0Fixed(k,s,t);
            floatingScaleConstraint.setP1Floating(k,s,t);
        }
        else if(!floatingScaleConstraint.p1fixed)
        {
            floatingScaleConstraint.setP1Fixed(k,s,t);
            floatingScaleConstraint.setP2Floating(k,s,t);
        }
        else
        {
            floatingScaleConstraint.setP2Fixed(k,s,t);
            finishScaleConstraint(true);
        }
    }
}

void IGLMesh::finishScaleConstraint(bool restartSmoothing)
{
    Vector3d p0 = getFlatSurfacePoint(floatingScaleConstraint.face0,floatingScaleConstraint.u0,floatingScaleConstraint.v0);
    Vector3d p1 = getFlatSurfacePoint(floatingScaleConstraint.face1,floatingScaleConstraint.u1,floatingScaleConstraint.v1);
    Vector3d pmid = (p0 + p1)/2;

    Vector3d nmid = Vector3d(0.5,0.5,0.5);
    nmid.normalize();

    Vector3d D = -nmid;

    pmid = pmid + nmid;

    double t = 0;
    int k = 0;
    double u = 0;
    double v = 0;

    intersectMeshWithRayFlat(pmid,D,&t,&k,&u,&v);
    if(k >= 0)
    {
        floatingScaleConstraint.setMidPoint(k,u,v);
        computeSingleScaleConstraintScale(&floatingScaleConstraint);
        scaleConstraints2.push_back(floatingScaleConstraint);
        floatingScaleConstraint.clear();
        int sc_id = scaleConstraints2.size() - 1;
        perFaceScaleConstraintList[k].push_back(sc_id);

        //Vector3d p0 = getWrinkledSurfacePoint(floatingScaleConstraint.face0,floatingScaleConstraint.u0,floatingScaleConstraint.v0);
        //Vector3d p1 = getWrinkledSurfacePoint(floatingScaleConstraint.face1,floatingScaleConstraint.u1,floatingScaleConstraint.v1);
        //double scaleY = (p1-p0).norm();

        gatherScales(k);
    }
    needsFlattening = true;
    emit autoSave();
    if(restartSmoothing)
    {
        allowSmoothing();
    }
}

void IGLMesh::computeSingleScaleConstraintScale(ScaleConstraint *sc)
{
    Vector3d p0 = getWrinkledSurfacePoint(sc->face0,sc->u0,sc->v0);
    Vector3d p1 = getWrinkledSurfacePoint(sc->face1,sc->u1,sc->v1);
    double scaleY = (p1-p0).norm();
    double scaleX = scaleY;
    if(sc->p2fixed)
    {
        Vector3d p2 = getWrinkledSurfacePoint(sc->face2,sc->u2,sc->v2);
        Vector3d pmid = getWrinkledSurfacePoint(sc->face,sc->u,sc->v);
        scaleX = (p2-pmid).norm();
    }
    sc->scale(0) = scaleX;
    sc->scale(1) = scaleY;
}

void IGLMesh::beginSmoothing()
{
    QThread* thread = new QThread;
    SmoothingWorker* worker = new SmoothingWorker(this);
    worker->moveToThread(thread);

    connect(thread, SIGNAL(started()), worker, SLOT(process()));
    connect(worker, SIGNAL(finished()), thread, SLOT(quit()));
    connect(worker, SIGNAL(finished()), worker, SLOT(deleteLater()));
    connect(thread, SIGNAL(finished()), thread, SLOT(deleteLater()));

    connect(worker, SIGNAL(finished()), this, SLOT(workerFinished()));
    //connect(thread, SIGNAL(finished()), this, SLOT(workerFinished()));
    thread->start();
}

void IGLMesh::allowSmoothing()
{
    qDebug() << "allow smoothing";
    smoothingAllowed = true;
}

void IGLMesh::disallowSmoothing()
{
    smoothingAllowed = false;
    flatteningEnabled = false;
}

QVector<int> IGLMesh::containsInvertedElements()
{
    QVector<int> result;
    isInverted.clear();
    for(int i = 0; i < F.cols(); ++i)
    {
        int v0 = F(0,i);
        int v1 = F(1,i);
        int v2 = F(2,i);

        Vector3d p0 = flatV.col(v0);
        Vector3d p1 = flatV.col(v1);
        Vector3d p2 = flatV.col(v2);

        Vector3d p01 = p1 - p0;
        Vector3d p02 = p2 - p0;

        Vector3d N = p01.cross(p02);
        N.normalize();

        if(N(2) < 0)
        {
            result.push_back(i);
            qDebug() << "Inverted element at: " << i;
            isInverted.push_back(true);
        }
        else
        {
            isInverted.push_back(false);
        }
    }
    return result;
}

QVector<int> IGLMesh::containsInvertedElements(MatrixXd Vs)
{
    QVector<int> result;
    isInverted.clear();
    for(int i = 0; i < F.cols(); ++i)
    {
        int v0 = F(0,i);
        int v1 = F(1,i);
        int v2 = F(2,i);

        Vector3d p0 = Vs.col(v0);
        Vector3d p1 = Vs.col(v1);
        Vector3d p2 = Vs.col(v2);

        Vector3d p01 = p1 - p0;
        Vector3d p02 = p2 - p0;

        Vector3d N = p01.cross(p02);
        N.normalize();

        if(N(2) < 0)
        {
            result.push_back(i);
            //qDebug() << "Inverted element at: " << i;
            isInverted.push_back(true);
        }
        else
        {
            isInverted.push_back(false);
        }
    }
    return result;
}

void IGLMesh::removeFace(int k)
{
    MatrixXi newF;
    MatrixXi newFTC;
    VectorXi newMat;
    newF.setZero(F.rows(),F.cols()-1);
    newFTC.setZero(FTC.rows()-1,FTC.cols());
    newMat.setZero(Materials.rows()-1);

    int j = 0;
    for(int i = 0; i < F.cols(); ++i)
    {
        if(i == k)
            continue;
        newF.col(j) = F.col(i);
        newFTC.row(j) = FTC.row(i);
        newMat.row(j) = Materials.row(i);
        ++j;
    }
    F = newF;
    FTC = newFTC;
    Materials = newMat;
}

SmoothingWorker::SmoothingWorker(IGLMesh *_mesh) : mesh(_mesh)
{
}

SmoothingWorker::~SmoothingWorker()
{
}

void SmoothingWorker::process()
{
    mesh->doAllSmoothing();
    //FlatteningEngine::rebuildLimData(mesh);
    emit finished();
}

void IGLMesh::rotate3DMeshAroundZBy(double theta)
{
    Matrix3d m;
    AngleAxisd aad = AngleAxisd(theta * M_PI / 180.0, Vector3d::UnitZ());
    m = aad.toRotationMatrix();
    for(int i = 0; i < V.cols(); ++i)
    {
        Vector3d oldv = V.col(i);
        Vector3d newv = m * oldv;
        V.col(i) = newv;
    }
}

MatrixXd IGLMesh::interpolate3Dto2D(double t)
{
    if(t <= 0.0)
        return V;
    if(t >= 1.0)
        return flatV;

    MatrixXd Vinterp = V;
    for(int i = 0; i < V.cols(); ++i)
    {
        Vinterp.col(i) = t * flatV.col(i) + (1.0 - t) * V.col(i);
    }
    return Vinterp;
}

double IGLMesh::getFaceArea(int k)
{
    int v0 = F(0,k);
    int v1 = F(1,k);
    int v2 = F(2,k);

    Vector3d wv0 = V.col(v0);
    Vector3d wv1 = V.col(v1);
    Vector3d wv2 = V.col(v2);

    Vector3d wv01 = wv1 - wv0;
    Vector3d wv02 = wv2 - wv0;

    double area = (wv01.cross(wv02)).norm() / 2.0;

    return area;
}

double IGLMesh::getFlatFaceArea(int k)
{
    int v0 = F(0,k);
    int v1 = F(1,k);
    int v2 = F(2,k);

    Vector3d fv0 = flatV.col(v0);
    Vector3d fv1 = flatV.col(v1);
    Vector3d fv2 = flatV.col(v2);

    Vector3d fv01 = fv1 - fv0;
    Vector3d fv02 = fv2 - fv0;

    double area = (fv01.cross(fv02)).norm() / 2.0;

    return area;
}

bool IGLMesh::containsUnreferecedVertices()
{
    QVector<bool> isReferenced = QVector<bool>(V.cols(),false);

    for(int k = 0; k < F.cols(); ++k)
    {
        int v0 = F(0,k);
        int v1 = F(1,k);
        int v2 = F(2,k);

        isReferenced[v0] = true;
        isReferenced[v1] = true;
        isReferenced[v2] = true;
    }

    for(int k = 0; k < isReferenced.size(); ++k)
    {
        if(!isReferenced[k])
        {
            return true;
        }
    }

    return false;
}

bool IGLMesh::containsStrayComponents()
{
    //2      create a queue Q
    QQueue<int> vertexFringe;

    //3      create a vector set V
    QVector<int> componentLabel = QVector<int>(V.cols(),-1);
    int currentLabel = 0;

    for(int k = 0; k < V.cols(); ++k)
    {
        if(componentLabel[k] == -1)
        {
            currentLabel++;
            vertexFringe.clear();
            //4      enqueue v onto Q
            vertexFringe.enqueue(k);

            //5      add v to V
            componentLabel[k] = currentLabel;

            //6      while Q is not empty loop
            while(!vertexFringe.isEmpty())
            {
                //7         t ← Q.dequeue()
                int t = vertexFringe.dequeue();
                QSet<int> t_neighbours;
                for(int i = 0; i < incidentFaces[t].size(); ++i)
                {
                    int v0 = F(0,incidentFaces[t][i]);
                    int v1 = F(1,incidentFaces[t][i]);
                    int v2 = F(2,incidentFaces[t][i]);

                    t_neighbours.insert(v0);
                    t_neighbours.insert(v1);
                    t_neighbours.insert(v2);
                }
                t_neighbours.remove(t);
                QVector<int> t_neighboursvec = t_neighbours.toList().toVector();
                for(int i = 0; i < t_neighboursvec.size(); ++i)
                {
                    //12           u ← G.adjacentVertex(t,e)
                    int u = t_neighboursvec[i];
                    //13           if u is not in V then
                    if(componentLabel[u] == -1)
                    {
                        //14               add u to V
                        componentLabel[u] = currentLabel;
                        //15               enqueue u onto Q
                        vertexFringe.enqueue(u);
                    }
                }
            }
        }
    }
    if(currentLabel > 1)
        return true;
    else
        return false;
}
