//
//  IGLMesh.h
//  ParchmentFlattener
//
//  Copyright (c) 2014 Kazim Pal.
//  Released under the MIT License.
//
#pragma once
#include <Eigen/Sparse>
#include <Eigen/Dense>
#include <qstring.h>
#include <QVector>
#include <QPair>
#include <qvector3d.h>
#include <qstringlist.h>
#include <QtConcurrentRun>

#define VECTOR_SMOOTH_TOLERANCE 0.05

using namespace Eigen;

class AABBTree;
class FlatteningEngine;
class IGLMesh;

class LineConstraint
{
public:
    LineConstraint(void);
    ~LineConstraint(void);

    MatrixXd sts;
    QVector<int> faces;

    MatrixXd new_sts;
    QVector<int> new_faces;

    MatrixXd suggested_sts;
    QVector<int> suggested_faces;

    MatrixXd rightVectorsW;
    MatrixXd upVectorsW;
    MatrixXd rightVectorsF;
    MatrixXd upVectorsF;

    void addPoint(int k, double s, double t);
    int size();
    void clear();

    void addSuggestionPoint(int k, double s, double t);
    void clearSuggestions();
    void acceptSuggestions();
    void acceptSuggestions(int n);

    bool isActive();
    void deleteLineConstraint();

    bool mIsVertical;

private:
    bool active;
};

class ScaleConstraint
{
public:
    ScaleConstraint(void);
    ~ScaleConstraint(void);

    int face;
    double u;
    double v;

    int face0;
    double u0,v0;

    int face1;
    double u1,v1;

    int face2;
    double u2, v2;

    Vector2d scale;
    double adjustmentFactor;
    bool p0fixed;
    bool p1fixed;
    bool p2fixed;

    void setP0Floating(int k, double u, double v);
    void setP0Fixed(int k, double u, double v);
    void setP1Floating(int k, double u, double v);
    void setP1Fixed(int k, double u, double v);
    void setP2Floating(int k, double u, double v);
    void setP2Fixed(int k, double u, double v);
    void setMidPoint(int _face, double _u, double _v);
    void clear();

    bool isActive();
    void deleteScaleConstraint();

private:
    bool active;
};

QDataStream &operator>>(QDataStream &in, ScaleConstraint &sc);

class SmoothingWorker : public QObject
{
Q_OBJECT
public:
    SmoothingWorker(IGLMesh *_mesh);
    ~SmoothingWorker();

public slots:
    void process();

signals:
    void finished();
    void error(QString err);

private:
    IGLMesh *mesh;
};

class IGLMesh : public QObject
{
Q_OBJECT
public:
    IGLMesh(void);
    ~IGLMesh(void);

	MatrixXd V;
    MatrixXd Vuntouched;
    MatrixXd Vinitial;
    MatrixXd flatV;
	MatrixXi F;
	MatrixXd VN;
    MatrixXd TN;
	MatrixXi FN;
	MatrixXd TC;
	MatrixXi FTC;
	VectorXi Materials;
    Vector3d C;
    Vector3d flatC;

    QVector< QVector<int> > incidentFaces;
    QVector< QVector<int> > perFaceScaleConstraintList;
    QVector< QVector< QPair<int,int> > > perFaceVertexConstraintList;
    QStringList texFNames;

    QString meshName;

    QVector<bool> isInverted;

    Vector3d aveNormal;
    Vector3d aveRightVectorF;

    MatrixXd perVertexRightVectorsW;
    MatrixXd perVertexUpVectorsW;

    MatrixXd perVertexRightVectorsF;
    MatrixXd perVertexUpVectorsF;

    MatrixXd perFaceRightVectorsW;
    MatrixXd perFaceUpVectorsW;

    MatrixXd perFaceRightVectorsF;
    MatrixXd perFaceUpVectorsF;

    Matrix2Xd perVertexScaleW;
    Matrix2Xd perVertexScaleF;
    Matrix2Xd perFaceScaleW;
    Matrix2Xd perFaceScaleF;

    double getDiffusionWeightBetweenVertices(int v0, int v1);
    //double smoothVectorsAndScalesSingleIteration();
    //double smoothVectorsAndScalesSingleIterationFlat();

    //void computeAllPairsGeodesicDistances();
    void computeVectorAndScalesDiffusion();
    //void computeVectorsAndScalesRadialBasisGeodesic();
    //void computeVectorsAndScalesRadialBasisNaive();
    void transferVertexGradientToFaces3D();
    //void transferVertexGradientToFaces2D();
    void transferCrossFieldFrom3Dto2D();
    void smooth2DCrossField();

    bool isAligned;
    void alignAveRightVecWithXAxis();

    void doAllSmoothing();

    void computeNormals();
    void getIncidentFaces();

    MatrixXd vectorConstraints;
    VectorXi isVectorConstrained;
    VectorXd scaleConstraints;
    VectorXi isScaleConstrained;
    VectorXi isFaceVectorConstrained;

    void clearConstraints();

    QVector<LineConstraint> lineConstraints;
    LineConstraint floatingLineConstraint;
    void clickLineConstraint(int k, double s, double t);
    void finishLineConstraint(bool isVertical);
    void deleteCurrentLineConstraint();
    int selectedLineConstraint;
    int selectedLineConstraintPoint;
    void moveSelectedLineConstraintPoint(int k, double s, double t);
    void computeSingleLineConstraintRightAndUpVectors(LineConstraint *lc);
    void computeAllLineConstraintRightAndUpVectors();
    bool aLineConstraintWasMoved;
    void finishEditingLineConstraint();

    QVector<ScaleConstraint> scaleConstraints2;
    ScaleConstraint floatingScaleConstraint;
    void clickScaleConstraint(int k, double s, double t, bool rightClick);
    void finishScaleConstraint(bool restartSmoothing);
    int selectedScaleConstraint;
    void deleteCurrentScaleConstraint();
    void computeSingleScaleConstraintScale(ScaleConstraint *sc);

    void selectedLineConstraintForward();
    void selectedScaleConstraintForward();
    void selectedLineConstraintBackward();
    void selectedScaleConstraintBackward();

    void addVectorConstraint(int i, double vx, double vy, double vz);
    void addScaleConstraint(int i, double r);

	void loadMesh(QString filename);
    void loadTextureCoords(QString filename);
    void centerMesh();
    void orientMesh();
    void render();

    void saveFlatMesh(QString filename);

    void saveFlatVertices(QString filename);
    void loadFlatVertices(QString filename);

    void loadConstraints(QString filename);
    void saveConstraints(QString filename);
    void saveConstraintsAbs(QString filename);

    void setVertexPositions(MatrixXd newV);
    void setFlatVertexPositions(MatrixXd newV);
    void setInitialVertexPositions(MatrixXd initV);
    int selected_vertex;
    void unSelectVertex();
    
    QVector3D getQVector3D(int i);
    QVector3D getQVector3DNormal(int i);

    bool flattened;

    Vector3d getWrinkledSurfacePoint(int k, double s, double t);
    Vector3d getFlatSurfacePoint(int k, double s, double t);
    Vector3d getInitSurfacePoint(int k, double s, double t);
    Vector3d getSurfaceNormal(int k, double u, double v);

    int hover_k;
    double hover_u, hover_v;
    bool isHover;
    Vector3d fake_hover;

    bool intersectMeshWithRay(Vector3d O, Vector3d D, double *t, int *k, double *u, double *v);
    bool intersectMeshWithRayFlat(Vector3d O, Vector3d D, double *t, int *k, double *u, double *v);
    bool intersectTriangleWithRay(Vector3d O, Vector3d D, Vector3d vert0, Vector3d vert1, Vector3d vert2, double *t, double *u, double *v);

    AABBTree *aabbtree;

    void gatherVectors(int f_id);
    void gatherVectorsFlat(int f_id);
    void gatherScales(int f_id);
    void reGatherCurrentConstraints();

    void loadScaleConstraints(QString filename, double scoreThreshold);

    bool renderCrossField;

    QThread *smoothThread;
    bool smoothIsPaused;
    bool flatteningEnabled;
    bool flatteningLocked;

    void startSmoothing();

    void removeFace(int k);
    QVector<int> containsInvertedElements();
    QVector<int> containsInvertedElements(MatrixXd Vs);

    FlatteningEngine *flatteningEngine;

    MatrixXd interpolate3Dto2D(double t);
    void rotate3DMeshAroundZBy(double theta);

    double getFaceArea(int k);
    double getFlatFaceArea(int k);

    bool containsUnreferecedVertices();
    bool containsStrayComponents();

public slots:
    void allowSmoothing();
    void disallowSmoothing();
    void workerFinished();


signals:
    void smoothingIterationFinished();
    void flatteningIterationFinished();
    void emitNewTS(int newts);
    void autoSave();

private:
    bool loadMaterials(QString objFilename, QString mtlFilename);
    bool smoothingAllowed;
    void beginSmoothing();

    bool needsSmoothing;
    bool needsFlattening;
};

